# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for Autopkgtest work request plugin."""

import textwrap
from typing import ClassVar, override
from unittest.mock import patch

from django.http import HttpResponseBase
from django.template.loader import get_template
from django.template.response import SimpleTemplateResponse

from debusine.artifacts import AutopkgtestArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianAutopkgtest,
    DebianAutopkgtestSource,
)
from debusine.db.models import Artifact, WorkRequest, Workspace
from debusine.db.playground import scenarios
from debusine.tasks.models import (
    AutopkgtestData,
    AutopkgtestDynamicData,
    AutopkgtestInput,
    BackendType,
    LookupMultiple,
    LookupSingle,
)
from debusine.test.django import TestCase
from debusine.web.views import files
from debusine.web.views.autopkgtest import (
    AutopkgtestLogParser,
    AutopkgtestTab,
    AutopkgtestViewWorkRequestPlugin,
)
from debusine.web.views.files import LogFileWidget, TooBigFileWidget
from debusine.web.views.work_request import Tab


class AutopkgtestViewWorkRequestPluginTests(TestCase):
    """Tests for AutopkgtestViewWorkRequestPlugin."""

    environment_artifact: ClassVar[Artifact]
    source_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.environment_artifact, _ = cls.playground.create_artifact(
            paths=["system.tar.xz"],
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={
                "vendor": "debian",
                "codename": "bookworm",
                "architecture": "amd64",
                "variant": "",
            },
            create_files=True,
        )
        cls.source_artifact, _ = cls.playground.create_artifact(
            paths=["hello.dsc"],
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={"name": "hello", "version": "1.0", "type": "dpkg"},
            create_files=True,
        )

    def get_tab(self, work_request: WorkRequest) -> AutopkgtestTab:
        """Return the tab for a work request or a plugin."""
        plugin = AutopkgtestViewWorkRequestPlugin(work_request.get_task())
        tabs: list[Tab] = []
        plugin.add_tabs(tabs)
        self.assertEqual(len(tabs), 1)
        tab = tabs[0]
        assert isinstance(tab, AutopkgtestTab)
        self.assertEqual(tab.name, "autopkgtest")
        self.assertEqual(tab.label, "Autopkgtest")
        return tab

    def create_work_request_for_artifacts(
        self,
        environment_artifact: Artifact,
        source_artifact: Artifact,
        binary_artifacts: list[Artifact],
        context_artifacts: list[Artifact] | None = None,
        workspace: Workspace | None = None,
    ) -> WorkRequest:
        """Create an autopkgtest work request for some artifacts."""
        environment_lookup: int | str
        if workspace is None:
            workspace = self.playground.get_default_workspace()
        if "vendor" in environment_artifact.data:
            environments_collection = self.playground.create_collection(
                environment_artifact.data["vendor"],
                CollectionCategory.ENVIRONMENTS,
                workspace=workspace,
            )
            environments_collection.manager.add_artifact(
                environment_artifact,
                user=self.playground.get_default_user(),
                variables={"backend": BackendType.UNSHARE},
            )
            environment_lookup = "{vendor}/match:codename={codename}".format(
                **environment_artifact.data
            )
        else:
            environment_lookup = environment_artifact.id
        task_data_input = AutopkgtestInput(
            source_artifact=source_artifact.id,
            binary_artifacts=LookupMultiple.model_validate(
                [binary_artifact.id for binary_artifact in binary_artifacts]
            ),
        )
        if context_artifacts:
            task_data_input.context_artifacts = LookupMultiple.model_validate(
                [context_artifact.id for context_artifact in context_artifacts]
            )

        return self.playground.create_worker_task(
            result=WorkRequest.Results.SUCCESS,
            task_data=AutopkgtestData(
                build_architecture="amd64",
                environment=environment_lookup,
                input=task_data_input,
            ),
            task_name="autopkgtest",
            workspace=workspace,
            use_data=False,
        )

    def create_autopkgtest_artifact(
        self, work_request: WorkRequest
    ) -> Artifact:
        """Create an AutopkgtestArtifact for a work request."""
        artifact, _ = self.playground.create_artifact(
            paths=["log", "summary"],
            work_request=work_request,
            create_files=True,
        )
        artifact.category = AutopkgtestArtifact._category
        artifact.data = {
            "results": {
                "command1": {"status": "PASS"},
                "command2": {"status": "FAIL", "details": "partial"},
            },
            "cmdline": "/usr/bin/hello",
            "source_package": {
                "name": "hello",
                "version": "1.0",
                "url": "https://example.org/hello.deb",
            },
            "architecture": "amd64",
            "distribution": "debian:sid",
        }
        artifact.save()
        return artifact

    def test_is_enabled(self) -> None:
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        autopkgtest_view = AutopkgtestViewWorkRequestPlugin(
            work_request.get_task()
        )
        self.assertTrue(autopkgtest_view.is_enabled())

    def test_get_context_data_simple(self) -> None:
        """get_context_data returns the expected data for a simple case."""
        binary_artifact_1, _ = self.playground.create_artifact(
            paths=["hello-1.deb"], create_files=True
        )
        binary_artifact_2, _ = self.playground.create_artifact(
            paths=["hello-2.deb"], create_files=True
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact,
            self.source_artifact,
            [binary_artifact_1, binary_artifact_2],
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[
                binary_artifact_1.id,
                binary_artifact_2.id,
            ],
        ).model_dump()
        work_request.save()
        result_artifact = self.create_autopkgtest_artifact(work_request)
        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data,
            {
                "request_data": {
                    "backend": BackendType.AUTO,
                    "binary_artifacts": {
                        "lookup": [binary_artifact_1.id, binary_artifact_2.id],
                        "artifacts": [
                            {
                                "files": ["hello-1.deb"],
                                "id": binary_artifact_1.id,
                                "artifact": binary_artifact_1,
                            },
                            {
                                "files": ["hello-2.deb"],
                                "id": binary_artifact_2.id,
                                "artifact": binary_artifact_2,
                            },
                        ],
                    },
                    "build_architecture": "amd64",
                    "codename": "bookworm",
                    "context_artifacts": {"lookup": [], "artifacts": []},
                    "debug_level": 0,
                    "exclude_tests": [],
                    "extra_repositories": None,
                    "extra_environment": {},
                    "fail_on_scenarios": ["failed"],
                    "include_tests": [],
                    "needs_internet": "run",
                    "source_artifact": {
                        "files": ["hello.dsc"],
                        "id": self.source_artifact.id,
                        "lookup": self.source_artifact.id,
                        "artifact": self.source_artifact,
                    },
                    "source_name": "hello",
                    "source_version": "1.0",
                    "task_description": "hello_1.0_amd64 in debian:bookworm",
                    "timeout": None,
                    "use_packages_from_base_repository": False,
                    "vendor": "debian",
                },
                "result": work_request.result,
                "result_artifact": result_artifact,
                "result_data": [
                    ("command1", "PASS", None),
                    ("command2", "FAIL", "partial"),
                ],
            },
        )

    def test_get_context_data_with_context_artifacts(self) -> None:
        """Any context artifacts show up in the context data."""
        context_artifact_1, _ = self.playground.create_artifact(
            paths=["hello-rdep-1.deb"], create_files=True
        )
        context_artifact_2, _ = self.playground.create_artifact(
            paths=["hello-rdep-2.deb"], create_files=True
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact,
            self.source_artifact,
            [],
            context_artifacts=[context_artifact_1, context_artifact_2],
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
            input_context_artifacts_ids=[
                context_artifact_1.id,
                context_artifact_2.id,
            ],
        ).model_dump()
        work_request.save()
        context_data = self.get_tab(work_request).get_context_data()

        self.assertEqual(
            context_data["request_data"]["context_artifacts"],
            {
                "lookup": [context_artifact_1.id, context_artifact_2.id],
                "artifacts": [
                    {
                        "files": ["hello-rdep-1.deb"],
                        "id": context_artifact_1.id,
                    },
                    {
                        "files": ["hello-rdep-2.deb"],
                        "id": context_artifact_2.id,
                    },
                ],
            },
        )

    def test_context_data_no_artifacts(self) -> None:
        """If there are no output artifacts, context data omits that info."""
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        context_data = self.get_tab(work_request).get_context_data()
        self.assertCountEqual(context_data.keys(), {"request_data", "result"})

    def test_context_data_no_dynamic_task_data(self) -> None:
        """If there is no dynamic data, context data omits artifact info."""
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        context_data = self.get_tab(work_request).get_context_data()
        self.assertIsNone(context_data["request_data"]["codename"])
        self.assertIsNone(context_data["request_data"]["vendor"])
        self.assertEqual(
            context_data["request_data"]["task_description"],
            "UNKNOWN_UNKNOWN_amd64",
        )
        self.assertEqual(
            context_data["request_data"]["source_artifact"],
            {
                "lookup": self.source_artifact.id,
                "id": None,
                "files": [],
                "artifact": None,
            },
        )

    def test_context_data_missing_environment_artifact(self) -> None:
        """If the environment artifact is missing, context data omits it."""
        environment_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()
        environment_artifact.delete()
        context_data = self.get_tab(work_request).get_context_data()
        self.assertIsNone(context_data["request_data"]["codename"])
        self.assertIsNone(context_data["request_data"]["vendor"])
        self.assertEqual(
            context_data["request_data"]["task_description"], "hello_1.0_amd64"
        )

    def test_context_data_environment_artifact_wrong_category(self) -> None:
        """We cope with the environment artifact having an odd category."""
        environment_artifact, _ = self.playground.create_artifact(
            paths=["system.tar.xz"],
            category=ArtifactCategory.TEST,
            create_files=True,
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()
        tab = self.get_tab(work_request)
        context_data = tab.get_context_data()
        self.assertIsNone(context_data["request_data"]["codename"])
        self.assertIsNone(context_data["request_data"]["vendor"])
        self.assertEqual(
            context_data["request_data"]["task_description"], "hello_1.0_amd64"
        )

    def test_context_data_missing_source_artifact(self) -> None:
        """If the source artifact is missing, context data omits its files."""
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()
        source_artifact_id = self.source_artifact.id
        self.source_artifact.files.clear()
        self.source_artifact.delete()
        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["source_artifact"],
            {
                "lookup": source_artifact_id,
                "id": source_artifact_id,
                "artifact": None,
                "files": [],
            },
        )

    def test_context_data_source_artifact_wrong_category(self) -> None:
        """We cope with the source artifact having an odd category."""
        self.source_artifact.category = ArtifactCategory.TEST
        self.source_artifact.save()
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()
        context_data = self.get_tab(work_request).get_context_data()
        self.assertIsNone(context_data["request_data"]["source_name"])
        self.assertIsNone(context_data["request_data"]["source_version"])
        self.assertEqual(
            context_data["request_data"]["task_description"],
            "UNKNOWN_UNKNOWN_amd64 in debian:bookworm",
        )

    def test_context_data_missing_binary_artifact(self) -> None:
        """If a binary artifact is missing, context data omits its files."""
        binary_artifact_1, _ = self.playground.create_artifact(
            paths=["hello-1.deb"], create_files=True
        )
        binary_artifact_2, _ = self.playground.create_artifact(
            paths=["hello-2.deb"], create_files=True
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[
                binary_artifact_1.id,
                binary_artifact_2.id,
            ],
        ).model_dump()
        work_request.save()
        binary_artifact_1_id = binary_artifact_1.id
        binary_artifact_1.files.clear()
        binary_artifact_1.delete()
        tab = self.get_tab(work_request)
        context_data = tab.get_context_data()
        self.assertEqual(
            context_data["request_data"]["binary_artifacts"],
            {
                "lookup": work_request.task_data["input"]["binary_artifacts"],
                "artifacts": [
                    {
                        "id": binary_artifact_1_id,
                        "files": [],
                        "artifact": None,
                    },
                    {
                        "id": binary_artifact_2.id,
                        "files": ["hello-2.deb"],
                        "artifact": binary_artifact_2,
                    },
                ],
            },
        )

    def test_get_context_data_skips_incomplete_files(self) -> None:
        """get_context_data() skips incomplete files."""
        source_artifact, _ = self.playground.create_artifact(
            paths=["hello.dsc"],
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={"name": "hello", "version": "1.0", "type": "dpkg"},
            create_files=True,
            skip_add_files_in_store=True,
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()

        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["source_artifact"]["files"], []
        )

    def assert_contains_artifact_information(
        self,
        response: HttpResponseBase,
        description: str,
        artifact_lookup: LookupSingle,
        artifact_id: int,
    ) -> None:
        """Assert contains information of an artifact: link and files."""
        artifact = Artifact.objects.get(id=artifact_id)
        files = artifact.fileinartifact_set.order_by("path").values_list(
            "path", flat=True
        )
        artifact_path = artifact.get_absolute_url()
        artifact_link = f'<a href="{artifact_path}">#{artifact_id}</a>'
        files_li = "".join([f"<li>{file}</li>" for file in files])
        self.assertContains(
            response,
            f"<li>{description} ({artifact_lookup}: {artifact_link})"
            f"<ul>{files_li}</ul>"
            f"</li>",
            html=True,
        )

    def assert_contains_artifacts_information(
        self,
        response: HttpResponseBase,
        description: str,
        artifact_lookup: LookupMultiple,
        artifact_ids: list[int],
    ) -> None:
        """Assert contains information of a set of artifacts."""
        expected_response = (
            f"<li>{description} ({artifact_lookup.export()})<ul>"
        )
        for artifact_id in artifact_ids:
            artifact = Artifact.objects.get(id=artifact_id)
            files = artifact.fileinartifact_set.order_by("path").values_list(
                "path", flat=True
            )
            artifact_path = artifact.get_absolute_url()
            artifact_link = f'<a href="{artifact_path}">#{artifact_id}</a>'
            files_li = "".join([f"<li>{file}</li>" for file in files])
            expected_response += f"<li>{artifact_link}<ul>{files_li}</ul></li>"
        expected_response += "</ul></li>"
        self.assertContains(response, expected_response, html=True)

    def test_template_output(self) -> None:
        """Generic output of the template."""
        workspace = self.playground.create_workspace(name="Public", public=True)
        environment = self.playground.create_debian_environment(
            workspace=workspace
        )
        assert environment.artifact
        source_artifact = self.playground.create_source_artifact(
            workspace=workspace
        )
        binary_artifact_1 = (
            self.playground.create_minimal_binary_package_artifact(
                package="hello-1"
            )
        )
        binary_artifact_2 = (
            self.playground.create_minimal_binary_package_artifact(
                package="hello-2"
            )
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            environment.artifact,
            source_artifact,
            [binary_artifact_1, binary_artifact_2],
            workspace=workspace,
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=environment.artifact.id,
            input_source_artifact_id=source_artifact.id,
            input_binary_artifacts_ids=[
                binary_artifact_1.id,
                binary_artifact_2.id,
            ],
        ).model_dump()
        work_request.save()
        self.create_autopkgtest_artifact(work_request)

        response = self.client.get(work_request.get_absolute_url())
        tree = self.assertResponseHTML(response, dump_on_error=True)
        title = self.assertHasElement(tree, "//head/title")
        self.assertTextContentEqual(
            title, f"debusine - Public - #{work_request.id}: autopkgtest"
        )

        result_output = get_template("web/_work_request-result.html").render(
            {"result": work_request.result}
        )
        self.assertContains(response, f"Result: {result_output}", html=True)

        self.assert_contains_artifact_information(
            response, "Source artifact", source_artifact.id, source_artifact.id
        )

        self.assert_contains_artifacts_information(
            response,
            "Binary artifacts",
            LookupMultiple((binary_artifact_1.id, binary_artifact_2.id)),
            [binary_artifact_1.id, binary_artifact_2.id],
        )

    def test_get_fail_on_scenarios(self) -> None:
        """_get_fail_on_scenarios processes the task's fail_on correctly."""
        sub_tests = [
            ({"failed_test": True}, ["failed"]),
            (
                {
                    "failed_test": False,
                    "flaky_test": True,
                    "skipped_test": False,
                },
                ["flaky"],
            ),
            (
                {
                    "failed_test": True,
                    "flaky_test": True,
                    "skipped_test": True,
                },
                ["failed", "flaky", "skipped"],
            ),
        ]

        for fail_on, expected_scenarios in sub_tests:
            with self.subTest(fail_on=fail_on):
                self.assertEqual(
                    AutopkgtestTab._get_fail_on_scenarios(fail_on),
                    expected_scenarios,
                )

    def test_show_configured_task_data(self) -> None:
        """The view shows configured task data if present."""
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )

        # Try unconfigured
        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["build_architecture"], "amd64"
        )

        # Try configured
        work_request.configured_task_data = work_request.task_data.copy()
        work_request.configured_task_data["build_architecture"] = "arm64"
        work_request.save()
        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["build_architecture"], "arm64"
        )


class AutopkgtestLogParserTests(TestCase):
    """Tests for AutopkgtestLogParser class."""

    scenario = scenarios.PackageBuildLog()

    def test_parse_log(self) -> None:
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        sections, number_of_lines = AutopkgtestLogParser(
            self.scenario.autopkgtest_log
        ).parse()
        self.assertEqual(
            number_of_lines, self.scenario.autopkgtest_log.count(b"\n")
        )

        self.assertEqual(
            sections,
            [
                Section(
                    text="Preparation",
                    color="primary",
                    subsections=[
                        SubSection(
                            text="start run",
                            color=None,
                            subsections=[],
                            line=1,
                            result=None,
                            expanded=True,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        ),
                        SubSection(
                            text="test bed setup",
                            color=None,
                            subsections=[],
                            line=6,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        ),
                        SubSection(
                            text="apt-source",
                            color=None,
                            subsections=[],
                            line=10,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        ),
                    ],
                ),
                Section(
                    text="Tests",
                    color="primary",
                    subsections=[
                        SubSection(
                            text="preparing testbed pybuild-autopkgtest",
                            color=None,
                            subsections=[],
                            line=15,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=False,
                            tooltip=None,
                        ),
                        SubSection(
                            text="pybuild-autopkgtest",
                            color="success",
                            subsections=[],
                            line=23,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=True,
                            tooltip="Pass",
                        ),
                        SubSection(
                            text="preparing testbed test-suite",
                            color=None,
                            subsections=[],
                            line=31,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=False,
                            tooltip=None,
                        ),
                        SubSection(
                            text="test-suite",
                            color="danger",
                            subsections=[],
                            line=43,
                            result=None,
                            expanded=True,
                            include_in_table_of_contents=True,
                            tooltip="Fail",
                        ),
                        SubSection(
                            text="preparing testbed test-suite-skip",
                            color=None,
                            subsections=[],
                            line=50,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=False,
                            tooltip=None,
                        ),
                        SubSection(
                            text="test-suite-skip",
                            color="secondary",
                            subsections=[],
                            line=51,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=True,
                            tooltip="Skip",
                        ),
                    ],
                ),
                Section(
                    text="Closing",
                    color="primary",
                    subsections=[
                        SubSection(
                            text="summary",
                            color=None,
                            subsections=[],
                            line=54,
                            result=None,
                            expanded=True,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        )
                    ],
                ),
            ],
        )

    def test_parse_log_without_apt_source(self) -> None:
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        log_contents = textwrap.dedent(
            """\
        line
        line
        15s autopkgtest [22:36:01]: test m-a-autopkgtest: preparing testbed
        line
        43s autopkgtest [22:36:29]: test m-a-autopkgtest: [-----------------------
        44s I: No Linux header packages are installed.
        line
        line
        """  # noqa: E501
        ).encode()

        sections, number_of_lines = AutopkgtestLogParser(log_contents).parse()
        self.assertEqual(number_of_lines, log_contents.count(b"\n"))
        self.assertEqual(len(sections), 3)
        self.assertEqual(sections[0].subsections, [])

        # 1 subsection: preparing and the test itself
        self.assertEqual(
            sections[1].subsections,
            [
                SubSection(
                    text="preparing testbed m-a-autopkgtest",
                    color=None,
                    subsections=[],
                    line=3,
                    result=None,
                    expanded=False,
                    include_in_table_of_contents=False,
                    tooltip=None,
                ),
                SubSection(
                    text="m-a-autopkgtest",
                    color=None,
                    subsections=[],
                    line=5,
                    result=None,
                    expanded=False,
                    include_in_table_of_contents=True,
                    tooltip=None,
                ),
            ],
        )
        self.assertEqual(sections[2].subsections, [])

    def test_parse_log_with_test_setup_error(self) -> None:
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        log_contents = textwrap.dedent(
            """\
            line
            line
            2s autopkgtest [00:00:01]: test unit-tests-tasks: preparing testbed
            3s E: erroneous package: Test dependencies are unsatisfiable.
            4s autopkgtest [00:00:01]: test unit-tests-sign: preparing testbed
            5s autopkgtest [20:30:43]: test unit-tests-sign: [-----------------------
            5s unit-tests-sign output
            5s autopkgtest [20:30:44]: test unit-tests-sign: -----------------------]
            5s autopkgtest [20:30:44]: test unit-tests-sign:  - - - - - - - - - - results - - - - - - - - - -
            5s unit-tests-sign             PASS
            5s autopkgtest [00:00:01]: @@@@@@@@@@@@@@@@@@@@ summary
            5s unit-tests-tasks  FAIL badpkg
            6s unit-tests-sign  PASS
            line
            line
            """  # noqa: E501
        ).encode()

        sections, _ = AutopkgtestLogParser(log_contents).parse()
        self.assertEqual(len(sections[1].subsections), 3)

        self.assertEqual(
            sections[1].subsections,
            [
                SubSection(
                    text="preparing testbed unit-tests-tasks",
                    color="danger",
                    subsections=[],
                    line=3,
                    result=None,
                    expanded=True,
                    include_in_table_of_contents=True,
                    tooltip="Fail",
                ),
                SubSection(
                    text="preparing testbed unit-tests-sign",
                    color=None,
                    subsections=[],
                    line=5,
                    result=None,
                    expanded=False,
                    include_in_table_of_contents=False,
                    tooltip=None,
                ),
                SubSection(
                    text="unit-tests-sign",
                    color="success",
                    subsections=[],
                    line=6,
                    result=None,
                    expanded=False,
                    include_in_table_of_contents=True,
                    tooltip="Pass",
                ),
            ],
        )

    def test_parse_log_inconsistent_log(self) -> None:
        """
        Log contains, in summary, a result for a test that has not been seen.

        In that case the summary test gets ignored.
        """
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        log_contents = textwrap.dedent(
            """\
            line
            line
            4s autopkgtest [00:00:01]: test unit-tests-sign: preparing testbed
            5s autopkgtest [20:30:43]: test unit-tests-sign: [-----------------------
            5s unit-tests-sign output
            5s autopkgtest [20:30:44]: test unit-tests-sign: -----------------------]
            5s autopkgtest [20:30:44]: test unit-tests-sign:  - - - - - - - - - - results - - - - - - - - - -
            5s unit-tests-sign             PASS
            5s autopkgtest [00:00:01]: @@@@@@@@@@@@@@@@@@@@ summary
            5s unit-tests-sign  PASS
            6s some-test-not-seen-before  PASS
            """  # noqa: E501
        ).encode()

        sections, _ = AutopkgtestLogParser(log_contents).parse()
        self.assertEqual(len(sections[1].subsections), 2)

        self.assertEqual(
            sections[1].subsections,
            [
                SubSection(
                    text="preparing testbed unit-tests-sign",
                    color=None,
                    subsections=[],
                    line=3,
                    result=None,
                    expanded=False,
                    include_in_table_of_contents=False,
                    tooltip=None,
                ),
                SubSection(
                    text="unit-tests-sign",
                    color="success",
                    subsections=[],
                    line=4,
                    result=None,
                    expanded=False,
                    include_in_table_of_contents=True,
                    tooltip="Pass",
                ),
            ],
        )

    def test_parse_log_with_error_before_any_test(self) -> None:
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        log_contents = textwrap.dedent(
            """\
            10s autopkgtest [10:30:23]: @@@@@@@@@@@@@@@@@@@@ apt-source
            15s autopkgtest [10:30:25]: ERROR: pkg fail.
            """
        ).encode()

        sections, _ = AutopkgtestLogParser(log_contents).parse()
        self.assertEqual(len(sections[1].subsections), 0)

    def test_parse_log_with_preparation_error(self) -> None:
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        log_contents = textwrap.dedent(
            """\
            1s autopkgtest [10:00:00]: starting date and time
            2s autopkgtest [10:00:01]: testbed dpkg architecture: amd64
            3s autopkgtest [10:00:02]: ERROR: "dpkg" failed with stderr
            4s Debug: line 1
            5s Debug: line 2
            6s autopkgtest [22:00:06]: test mytest: preparing testbed
            line
            """
        ).encode()

        sections, number_of_lines = AutopkgtestLogParser(log_contents).parse()
        self.assertEqual(
            sections[0].subsections[-1],
            SubSection(
                text="start run",
                color="danger",
                subsections=[],
                line=1,
                result=None,
                expanded=True,
                include_in_table_of_contents=True,
                tooltip=None,
            ),
        )
        self.assertEqual(number_of_lines, log_contents.count(b"\n"))
        self.assertEqual(sections[1].subsections[0].line, 6)

    def test_parse_log_with_early_preparation_error(self) -> None:
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        log_contents = textwrap.dedent(
            """\
            1s autopkgtest [21:00:00]: ERROR: "dpkg" failed with stderr
            """
        ).encode()

        sections, _ = AutopkgtestLogParser(log_contents).parse()
        self.assertEqual(len(sections[0].subsections), 0)


class AutopkgtestArtifactPluginTests(TestCase):
    """Tests for AutopkgtestArtifactPlugin class."""

    scenario = scenarios.PackageBuildLog()

    def test_get_context_data(self) -> None:
        artifact = self.scenario.autopkgtest_log_artifact
        log_file = artifact.fileinartifact_set.get(path="log")
        log_file.content_type = "text/plain; charset=utf-8"
        log_file.save()

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        assert isinstance(response, SimpleTemplateResponse)
        assert response.context_data is not None
        self.assertEqual(
            response.context_data["specialized_tab"],
            {
                "label": "Autopkgtest log",
                "slug": "autopkgtest-log",
                "template": "web/_autopkgtest-artifact-detail.html",
            },
        )

        (sections, _) = AutopkgtestLogParser(
            self.scenario.autopkgtest_log
        ).parse()

        self.assertEqual(response.context_data["sections"], sections)
        FileSection = LogFileWidget.Section

        self.assertEqual(
            response.context_data["log_widget"]._sections,
            [
                FileSection(
                    title="start run", start_line=1, end_line=5, expanded=True
                ),
                FileSection(
                    title="test bed setup",
                    start_line=6,
                    end_line=9,
                    expanded=False,
                ),
                FileSection(
                    title="apt-source",
                    start_line=10,
                    end_line=14,
                    expanded=False,
                ),
                FileSection(
                    title="preparing testbed pybuild-autopkgtest",
                    start_line=15,
                    end_line=22,
                    expanded=False,
                ),
                FileSection(
                    title="pybuild-autopkgtest",
                    start_line=23,
                    end_line=30,
                    expanded=False,
                ),
                FileSection(
                    title="preparing testbed test-suite",
                    start_line=31,
                    end_line=42,
                    expanded=False,
                ),
                FileSection(
                    title="test-suite",
                    start_line=43,
                    end_line=49,
                    expanded=True,
                ),
                FileSection(
                    title="preparing testbed test-suite-skip",
                    start_line=50,
                    end_line=50,
                    expanded=False,
                ),
                FileSection(
                    title="test-suite-skip",
                    start_line=51,
                    end_line=53,
                    expanded=False,
                ),
                FileSection(
                    title="summary", start_line=54, end_line=58, expanded=True
                ),
            ],
        )

        # Assert line number rendering is correct (has the ID and content)
        line_number_1 = self.assertHasElement(tree, "//a[@id='L1']")
        self.assertTextContentEqual(line_number_1, "1")

    def test_get_context_data_empty_log(self) -> None:
        data = DebianAutopkgtest(
            results={},
            cmdline="autopkgtest",
            source_package=DebianAutopkgtestSource(
                name="hello", version="1.0", url="https://example.com"
            ),
            architecture="amd64",
            distribution="debian:trixie",
        )

        artifact, _ = self.playground.create_artifact(
            paths={"log": b""},
            category=ArtifactCategory.AUTOPKGTEST,
            create_files=True,
            data=data,
        )
        log_file = artifact.fileinartifact_set.get(path="log")
        log_file.content_type = "text/plain; charset=utf-8"
        log_file.save()

        response = self.client.get(artifact.get_absolute_url())

        self.assertResponseHTML(response)

    def test_autopkgtest_without_log(self) -> None:
        data = DebianAutopkgtest(
            results={},
            cmdline="autopkgtest",
            source_package=DebianAutopkgtestSource(
                name="hello", version="1.0", url="https://example.com"
            ),
            architecture="amd64",
            distribution="debian:trixie",
        )

        artifact, _ = self.playground.create_artifact(
            paths={"log": b""},
            category=ArtifactCategory.AUTOPKGTEST,
            create_files=False,
            data=data,
        )

        response = self.client.get(artifact.get_absolute_url())
        tree = self.assertResponseHTML(response)

        error_paragraph = self.assertHasElement(
            tree, "//div[@id='error-artifact']"
        )

        self.assertTextContentEqual(
            error_paragraph,
            "Cannot display autopkgtest: Artifact does not contain log",
        )

    def test_autopkgtest_with_file_in_artifact_error(self) -> None:
        data = DebianAutopkgtest(
            results={},
            cmdline="autopkgtest",
            source_package=DebianAutopkgtestSource(
                name="hello", version="1.0", url="https://example.com"
            ),
            architecture="amd64",
            distribution="debian:trixie",
        )

        artifact, _ = self.playground.create_artifact(
            paths={"log": b""},
            category=ArtifactCategory.AUTOPKGTEST,
            create_files=True,
            skip_add_files_in_store=True,
            data=data,
        )

        response = self.client.get(artifact.get_absolute_url())
        tree = self.assertResponseHTML(response)

        error = self.assertHasElement(
            tree,
            "//div[@id='error-artifact']",
        )

        self.assertTextContentEqual(
            error,
            'Cannot display autopkgtest: '
            'File "log" has not completed uploading.',
        )

    @patch("debusine.web.views.files.MAX_FILE_SIZE", 10)
    def test_autopkgtest_with_file_too_big_to_render(self) -> None:
        data = DebianAutopkgtest(
            results={},
            cmdline="autopkgtest",
            source_package=DebianAutopkgtestSource(
                name="hello", version="1.0", url="https://example.com"
            ),
            architecture="amd64",
            distribution="debian:trixie",
        )

        artifact, _ = self.playground.create_artifact(
            paths={"log": b"x" * (files.MAX_FILE_SIZE + 1)},
            category=ArtifactCategory.AUTOPKGTEST,
            create_files=True,
            data=data,
        )

        response = self.client.get(artifact.get_absolute_url())
        self.assertResponseHTML(response)

        assert response.context_data is not None
        self.assertIsInstance(
            response.context_data["log_widget"], TooBigFileWidget
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Icon selection for debusine UI.

This module defines what icons are used in the UI to represent Debusine
concepts.
"""

import enum


class Icons(enum.StrEnum):
    """Maps constants to bootstrap5 icon names."""

    # Concepts
    ARTIFACT_RELATION_SOURCE = "arrow-left"
    ARTIFACT_RELATION_TARGET = "arrow-right"
    ARTIFACTS_INPUT = "box-arrow-in-down-right"
    ARTIFACTS_OUTPUT = "box-arrow-down-right"
    CATEGORY = "postcard"
    COLLECTION = "collection"
    COLLECTIONS = "stack"
    USER = "person-circle"
    WORKER = "pc-horizontal"
    WORKER_EXTERNAL = "pc-horizontal"
    WORKER_SIGNING = "patch-check"
    WORKFLOW = "diagram-3"
    WORK_REQUEST = "hammer"
    WORK_REQUEST_STATUS = "heart-pulse"
    WORK_REQUEST_SUPERSEDED = "arrow-repeat"
    WORK_REQUEST_SUPERSEDES = "shuffle"
    WORKSPACE = "box"
    SCOPE = "house"

    # Common attributes
    CREATED_AT = "clock"
    DURATION = "stopwatch"
    EXPIRE_AT = "hourglass-split"
    STARTED_AT = "rocket-takeoff"

    # Actions
    ARTIFACT_DOWNLOAD = "download"
    COLLECTION_ITEM_DETAILS = "link-45deg"
    FILE_DOWNLOAD = "file-earmark-arrow-down"
    FILE_VIEW = "file-earmark"
    FILE_VIEW_RAW = "file-earmark-code"
    WORK_REQUEST_RETRY = "arrow-repeat"
    WORK_REQUEST_ABORT = "x"
    DELETE = "trash"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Basic permission check infrastructure."""

import enum
import functools
import itertools
from collections.abc import Callable, Collection, Iterable
from typing import (
    Any,
    NamedTuple,
    Optional,
    Protocol,
    Self,
    TYPE_CHECKING,
    cast,
    runtime_checkable,
)

from django.core.exceptions import PermissionDenied
from django.db.models import Model, QuerySet
from django.utils.functional import classproperty

from debusine.db.permissioncontext import PermissionContext

if TYPE_CHECKING:
    from debusine.db.models import Scope, Workspace

type ImpliedByCollection = Collection[Any]


@runtime_checkable
class WithImplies[R: RoleBase](Protocol):
    """Define an object with an implies operator."""

    def implies(self, other: R) -> bool:
        """Implies operator."""


class Role(NamedTuple):
    """Annotated tuple used to define Roles."""

    value: str
    label: str | None = None
    implied_by: ImpliedByCollection = ()
    description: str | None = None


class RoleBase(str):
    """
    Base for role value implementations.

    This is intended to be subclassed and used as value for a role enum. For
    example::

        class ScopeRoleBase(RoleBase):
            ...

        class ScopeRoles(Roles, ScopeRoleBase, enum.ReprEnum):
            ...
    """

    def __new__(
        cls,
        value: str,
        label: str | None = None,  # noqa: ARG004, U100
        implied_by: ImpliedByCollection = (),  # noqa: ARG004, U100
        description: str | None = None,  # noqa: ARG004, U100
    ) -> Self:
        """Construct the string from value, ignoring other args."""
        return super().__new__(cls, value)

    def __init__(
        self,
        value: str,
        label: str | None = None,
        implied_by: ImpliedByCollection = (),
        description: str | None = None,
    ) -> None:
        """
        Set up RoleBase.

        :param value: string value for the role
        :param label: optional label (auto generated from value by default)
        :param implied_by: roles or other elements that can be used to imply
                           this role
        """
        self.label = label or value.capitalize()
        self.implied_by = implied_by
        self.description = description

    def _setup(self) -> None:
        """Set up the role after the enum is fully built."""

    @classmethod
    def from_iterable(cls, roles: Iterable[str]) -> frozenset[Self]:
        """Resolve a sequence of strings to a minimal set of ScopeRoles."""
        assert issubclass(cls, WithImplies)
        all_roles = frozenset(cls(str_role) for str_role in roles)
        if len(all_roles) == 1:
            return all_roles

        implied_roles = set()
        for a, b in itertools.permutations(all_roles, 2):
            if a != b and a.implies(b):
                implied_roles.add(b)
        return frozenset(all_roles - implied_roles)


# Mixin for role enums
class Roles:
    """Base for declaratively defining available roles for a resource."""

    @classmethod
    def setup(cls) -> None:
        """Finalize enum setup."""
        # Roles is used as a mixin for enum.ReprEnum, which is iterable,
        # but we cannot inherit from it without triggering the enum's setup.
        #
        # Overriding type checks until we have a better idea of how to type
        # this.
        for member in cls:  # type: ignore[attr-defined]
            member._setup()

    @classproperty
    def choices(cls) -> list[tuple[str, str]]:
        """Return labeled choices for use by Django fields."""
        res: list[tuple[str, str]] = []
        for entry in cls:  # type: ignore[attr-defined]
            role = cast("RoleBase", entry)
            res.append((str(role), role.label))
        return res


@runtime_checkable
class PermissionCheckPredicate[M: Model](Protocol):
    """Interface of a permission predicate on a resource."""

    __self__: M

    __func__: Callable[[M, PermissionContext], bool]

    def __call__(self, pc: PermissionContext) -> bool:
        """Test the predicate."""


def resolve_role(resource: Model | type[Model], role: str) -> str:
    """Convert a role in various forms into the right enum."""
    roles_enum = getattr(resource, "Roles")

    if isinstance(role, RoleBase):
        if not isinstance(role, roles_enum):
            raise TypeError(f"{role!r} cannot be converted to {roles_enum!r}")
        return cast(str, role)

    return cast(str, roles_enum(role))


def resolve_roles_list(
    resource: Model | type[Model], roles: str | Collection[str]
) -> list[str]:
    """Convert roles into a list of the right enum."""
    if not hasattr(resource, "Roles"):
        return []
    # Roles are string enums, which are instances of str, and therefore also of
    # Collection. This test intends to catch all roles that are not actual
    # collections of roles
    if isinstance(roles, str):
        return [resolve_role(resource, roles)]
    return [resolve_role(resource, r) for r in roles]


class Allow(enum.StrEnum):
    """
    Handling strategies for permission predicates.

    This is used to specify how to handle work_request and anonymous users.
    """

    #: The tested condition makes the predicate always fail
    NEVER = "never"
    #: The tested condition is ignored and the decision is delegated to the
    #: body of the predicate
    PASS = "pass"


type Predicate[M] = Callable[[M, PermissionContext], bool]


def permission_check[M: Model](
    msg: str,
    *,
    work_request: Allow = Allow.NEVER,
    anonymous: Allow = Allow.NEVER,
    explicit_grants_role: RoleBase | None = None,
) -> Callable[[Predicate[M]], Predicate[M]]:
    """
    Implement common elements of permission checking predicates.

    :param work_request: what to do if a work request token is passed
    :param anonymous: what to do if an anonymous user is passed
    :param explicit_grants_role: if a user token requiring explicit grants
      is passed, check for an explicit grant of this resource with this
      role; if such a token is passed and this parameter is unset, return
      False

    Note that a worker token currently implies that the user is anonymous. This
    may change with #523.

    Predicates should normally also check permissions on any containing
    resources, relying on query caching as needed for performance. This
    provides some defence in depth against omitted checks.
    """

    def wrap(f: Predicate[M]) -> Predicate[M]:
        @functools.wraps(f)
        def wrapper(self: M, pc: PermissionContext) -> bool:
            if pc.permission_checks_disabled:
                return True

            overrides = {
                Allow.NEVER: False,
                Allow.PASS: None,
            }
            if pc.work_request:
                if (override := overrides[work_request]) is not None:
                    return override
            elif pc.is_non_permission_bearing_token:
                return False
            elif pc.explicit_grants_only:
                if explicit_grants_role is not None:
                    # Delegate the decision of which resources should be
                    # allowed to per-model `extra_*_grants` querysets on the
                    # permission context, which should be derived from the
                    # token.
                    grants: QuerySet[Any] = getattr(
                        pc, f"extra_{self._meta.model.__name__.lower()}_grants"
                    )
                    assert hasattr(grants, "granted_resources")
                    granted_resources = grants.granted_resources(
                        role=explicit_grants_role
                    )
                    assert isinstance(granted_resources, QuerySet)
                    return granted_resources.filter(pk=self.pk).exists()
                else:
                    return False
            else:
                if not pc.user.is_authenticated:
                    if (override := overrides[anonymous]) is not None:
                        return override

            return f(self, pc)

        setattr(wrapper, "error_template", msg)
        setattr(wrapper, "work_request", work_request)
        setattr(wrapper, "anonymous", anonymous)
        setattr(wrapper, "explicit_grants_role", explicit_grants_role)
        return wrapper

    return wrap


type PredicateFilter[QS] = Callable[[QS, PermissionContext], QS]


def permission_filter[QS: QuerySet[Any, Any]](
    *,
    work_request: Allow = Allow.NEVER,
    anonymous: Allow = Allow.NEVER,
    explicit_grants_role: RoleBase | None = None,
) -> Callable[[PredicateFilter[QS]], PredicateFilter[QS]]:
    """
    Implement common elements of permission filtering predicates.

    :param work_request: what to do if a work request token is passed
    :param anonymous: what to do if an anonymous user is passed
    :param explicit_grants: what to do if a user token requiring explicit
      grants is passed
    :param explicit_grants_role: if a user token requiring explicit grants
      is passed, filter to resources explicitly granted with this role; if
      such a token is passed and this parameter is unset, return an empty
      queryset

    Note that a worker token currently implies that the user is anonymous. This
    may change with #523.

    Predicates should normally also check permissions on any containing
    resources, relying on query caching as needed for performance.  This
    provides some defence in depth against omitted checks.
    """

    def wrap(f: PredicateFilter[QS]) -> PredicateFilter[QS]:
        @functools.wraps(f)
        def wrapper(self: QS, pc: PermissionContext) -> QS:
            if pc.permission_checks_disabled:
                return self

            overrides = {
                Allow.NEVER: self.none(),
                Allow.PASS: None,
            }
            if pc.work_request:
                if (override := overrides[work_request]) is not None:
                    return override
            elif pc.is_non_permission_bearing_token:
                return self.none()
            elif pc.explicit_grants_only:
                if explicit_grants_role is not None:
                    # Delegate the decision of which resources should be
                    # allowed to per-model `extra_*_grants` querysets on the
                    # permission context, which should be derived from the
                    # token.
                    grants: QuerySet[Any] = getattr(
                        pc, f"extra_{self.model.__name__.lower()}_grants"
                    )
                    assert hasattr(grants, "granted_resources")
                    granted_resources = grants.granted_resources(
                        role=explicit_grants_role
                    )
                    assert isinstance(granted_resources, QuerySet)
                    return self.filter(pk__in=granted_resources)
                else:
                    return self.none()
            else:
                if not pc.user.is_authenticated:
                    if (override := overrides[anonymous]) is not None:
                        return override

            return f(self, pc)

        setattr(wrapper, "work_request", work_request)
        setattr(wrapper, "anonymous", anonymous)
        setattr(wrapper, "explicit_grants_role", explicit_grants_role)
        return wrapper

    return wrap


def format_permission_check_error(
    predicate: Callable[[PermissionContext], bool],
    pc: PermissionContext,
) -> str:
    """Format a permission check error message."""
    assert hasattr(predicate, "error_template")
    error_template = predicate.error_template
    assert isinstance(error_template, str)
    assert hasattr(predicate, "__self__")
    return error_template.format(
        resource=predicate.__self__, user=pc.user, pc=pc
    )


def enforce(predicate: Callable[[PermissionContext], bool]) -> None:
    """Enforce a permission predicate at the model level."""
    from debusine.db.context import context

    if predicate(context.pc):
        return

    raise PermissionDenied(
        format_permission_check_error(predicate, context.pc),
    )


def get_resource_scope(obj: Model) -> Optional["Scope"]:
    """
    Get the Scope object for a resource.

    :returns: None if the resource does not have a scope
    :raises NotImplementedError: for resources not currently supported
    """
    from debusine.db.models import ArtifactRelation, Scope, User

    match obj:
        case ArtifactRelation():
            return obj.artifact.workspace.scope
        case Scope():
            return obj
        case User():
            return None
        case _:
            if hasattr(obj, "scope"):
                assert isinstance(obj.scope, Scope)
                return obj.scope

            if hasattr(obj, "workspace"):
                assert isinstance(obj.workspace.scope, Scope)
                return obj.workspace.scope

            raise NotImplementedError(
                f"Cannot get scope for {obj.__class__.__name__} object"
            )


def get_resource_workspace[M: Model](obj: M) -> Optional["Workspace"]:
    """
    Get the Workspace object for a resource.

    :returns: None if the resource does not have a workspace
    :raises NotImplementedError: for resources not currently supported
    """
    from debusine.db.models import (
        ArtifactRelation,
        Group,
        Scope,
        User,
        Workspace,
    )

    match obj:
        case ArtifactRelation():
            return obj.artifact.workspace
        case Scope() | User() | Group():
            return None
        case Workspace():
            return obj
        case _:
            if hasattr(obj, "workspace"):
                assert isinstance(obj.workspace, Workspace)
                return obj.workspace

            raise NotImplementedError(
                f"Cannot get workspace for {obj.__class__.__name__} object"
            )


def anonymous_access(pc: PermissionContext) -> bool:  # noqa: ARG001
    """Permission predicate that grants access to all."""
    return True



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command manage_user."""

from typing import override
from unittest import mock

from django.contrib.auth import get_user_model
from django.core.management import BaseCommand, CommandError
from django.test import TestCase

from debusine.django.management.tests import call_command


class ManageUserCommandTests(TestCase):
    """Tests for manage_user management command."""

    @override
    def setUp(self) -> None:
        """Create a default User."""
        super().setUp()
        self.user = get_user_model().objects.create_user(
            username="bob", password="123456"
        )

    def test_no_action_raise_error(self) -> None:
        """Assert manage_user raise CommandError if no action is passed."""
        with self.assertRaises(CommandError):
            call_command("manage_user")

    def test_change_email_no_username_email_raise_error(self) -> None:
        """
        Assert manage_user change-email return an error message.

        Without the workaround in change_email.called_from_command_line
        it raises CommandError.
        """
        patcher = mock.patch.object(BaseCommand, "_called_from_command_line")
        mocked = patcher.start()
        self.addCleanup(patcher.stop)
        mocked.return_value = True

        _, _, exit_code = call_command("manage_user", "change-email")

        self.assertEqual(exit_code, 2)

    def test_change_email(self) -> None:
        """manage_user change-email changes the email."""
        email = "bob@bob.com"
        stdout, stderr, exit_code = call_command(
            "manage_user", "change-email", self.user.username, email
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.user.refresh_from_db()
        self.assertEqual(self.user.email, email)

    def test_change_email_username_not_found(self) -> None:
        """manage_user change-email user not found."""
        username = "non-existing-user"

        expected_error = f'Username "{username}" not found'
        with self.assertRaisesRegex(CommandError, expected_error) as exc:
            call_command(
                "manage_user",
                "change-email",
                username,
                "email@example.com",
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_change_email_username_system(self) -> None:
        """manage_user excludes system users."""
        expected_error = 'Username "_system" not found'
        with self.assertRaisesRegex(CommandError, expected_error) as exc:
            call_command(
                "manage_user", "change-email", "_system", "email@example.com"
            )

        self.assertEqual(exc.exception.returncode, 3)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Task to build Debian packages with sbuild.

This task implements the PackageBuild generic task for its task_data:
https://freexian-team.pages.debian.net/debusine/reference/tasks/ontology-generic-tasks.html#task-packagebuild
"""

import email.utils
import os
import re
import shlex
import shutil
import tempfile
from pathlib import Path
from typing import Any, cast, override

import pydantic
import yaml
from debian import debfile

import debusine.tasks.tags as ttags
import debusine.utils
import debusine.worker.tags as wtags
from debusine.artifacts import (
    BinaryPackage,
    DebugSymbols,
    PackageBuildLog,
    SigningInputArtifact,
    Upload,
)
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianSourcePackage,
    DebianSystemImage,
    DebianSystemTarball,
    DoseDistCheck,
    WorkRequestResults,
)
from debusine.client.models import RelationType, RemoteArtifact
from debusine.tasks import (
    BaseTaskWithExecutor,
    ExtraRepositoryMixin,
    RunCommandTask,
)
from debusine.tasks.inputs import (
    DebusineFQDNInput,
    EnvironmentInput,
    ExtraRepositoriesInput,
    MultiInput,
    SingleInput,
    Stage,
)
from debusine.tasks.models import (
    BackendType,
    SbuildBuildComponent,
    SbuildData,
    SbuildDynamicData,
)
from debusine.tasks.sbuild_validator_mixin import SbuildValidatorMixin
from debusine.utils import read_dsc

_BUILD_ID_DEBUG_PREFIX = "./usr/lib/debug/.build-id/"

_BUILD_ID_DEBUG_PATH_RE = re.compile(
    r"^\./usr/lib/debug/\.build-id/"
    r"(?P<prefix>[0-9a-f]{2})/(?P<suffix>[0-9a-f]{38})\.debug$"
)


class DebugSymbolPathError(Exception):
    """Raised when a -dbgsym .deb contains a malformed .build-id path."""


class Sbuild(
    SbuildValidatorMixin,
    ExtraRepositoryMixin[SbuildData, SbuildDynamicData],
    RunCommandTask[SbuildData, SbuildDynamicData],
    BaseTaskWithExecutor[SbuildData, SbuildDynamicData],
):
    """Task implementing a Debian package build with sbuild."""

    TASK_VERSION = 1

    input_source_artifact = SingleInput(
        field="input.source_artifact",
        categories=[ArtifactCategory.SOURCE_PACKAGE],
        stage=Stage.PENDING,
        verbose_name="Source package",
        help_text="Source artifact to build",
    )
    input_extra_binary_artifacts = MultiInput(
        field="input.extra_binary_artifacts",
        verbose_name="Extra binary artifacts",
        help_text="Extra packages to pass to sbuild --extra-package option",
    )
    environment = EnvironmentInput(
        categories=[
            ArtifactCategory.SYSTEM_TARBALL,
            ArtifactCategory.SYSTEM_IMAGE,
        ],
        stage=Stage.PENDING,
        verbose_name="Build environment",
        help_text="Tarball or system image used for the build",
    )
    extra_repositories = ExtraRepositoriesInput(
        help_text="Extra repositories used for the build"
    )
    debusine_fqdn = DebusineFQDNInput(
        verbose_name="Debusine FQDN",
        help_text="Domain of this Debusine instance",
        hidden=True,
    )

    @override
    def post_init(self) -> None:
        super().post_init()

        self.chroots: list[str] | None = None
        self.builder = "sbuild"

        # dsc_file Path. Set by self.fetch_input()
        self._dsc_file: Path | None = None
        self._extra_packages: list[Path] = []

        self._uploaded_binary_packages: list[
            tuple[BinaryPackage, RemoteArtifact]
        ] = []

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()
        assert isinstance(self.input_source_artifact.data, DebianSourcePackage)

        provided.add(
            ttags.ProvenanceProvided.SYSTEM,
            {
                ttags.SOURCE_PACKAGE_PREFIX
                + self.input_source_artifact.data.name
            },
        )

        required.add(
            wtags.ProvenanceRequired.SYSTEM, {wtags.CAP_PREFIX + "sbuild"}
        )

        return provided, required

    @override
    def get_subject(self) -> str | None:
        assert isinstance(self.input_source_artifact.data, DebianSourcePackage)
        return self.input_source_artifact.data.name

    @override
    def get_configuration_context(self) -> str | None:
        assert isinstance(
            self.environment.data, (DebianSystemTarball, DebianSystemImage)
        )
        return self.environment.data.codename

    @override
    def compute_dynamic_data(self) -> SbuildDynamicData:
        r"""
        Resolve artifact lookups for this task.

        :subject: package name of ``input.source_artifact``
        :parameter_summary: package name of
          ``input.source_artifact``\_"version" ("architecture")
        :runtime_context:
          ``$build_components:$build_architecture:$host_architecture`` or
          None if ``build_options`` is provided
        :configuration_context: codename of ``environment``
        """
        input_source_artifact = self.input_source_artifact
        assert isinstance(input_source_artifact.data, DebianSourcePackage)

        environment = self.environment
        assert isinstance(
            environment.data, (DebianSystemTarball, DebianSystemImage)
        )

        if self._build_options():
            # Do not record stats for non-standard builds
            runtime_context = None
        else:
            build_components = "+".join(sorted(self.data.build_components))

            runtime_context = (
                f"{build_components}:"
                f"{self.data.build_architecture}:"
                f"{self._host_architecture}"
            )

        source_package_name = input_source_artifact.data.name
        source_version = input_source_artifact.data.version

        return SbuildDynamicData(
            environment_id=environment.artifact_id,
            extra_repositories=self.extra_repositories,
            input_source_artifact_id=input_source_artifact.artifact_id,
            input_extra_binary_artifacts_ids=(
                self.input_extra_binary_artifacts.resolved_ids()
            ),
            binnmu_maintainer=f"Debusine <noreply@{self.debusine_fqdn}>",
            subject=self.get_subject(),
            parameter_summary=f"{source_package_name}_{source_version} "
            f"({self._host_architecture})",
            runtime_context=runtime_context,
            configuration_context=self.get_configuration_context(),
        )

    @property
    def _host_architecture(self) -> str:
        # Currently defining host_architecture for Sbuild is not
        # supported. It uses "build_architecture" (or "all" or "source")
        match self.data.build_components:
            case [SbuildBuildComponent.ALL]:
                return "all"
            case [SbuildBuildComponent.SOURCE]:
                return "source"
            case _:
                return self.data.build_architecture

    @override
    @classmethod
    def analyze_worker(cls) -> dict[str, Any]:
        """Report metadata for this task on this worker."""
        metadata = super().analyze_worker()

        available_key = cls.prefix_with_task_name("available")
        metadata[available_key] = debusine.utils.is_command_available("sbuild")

        return metadata

    @override
    @classmethod
    def get_provided_worker_tags(cls) -> set[str]:
        tags = super().get_provided_worker_tags()
        if debusine.utils.is_command_available("sbuild"):
            tags.add(wtags.CAP_PREFIX + "sbuild")
        return tags

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """Check the specified worker can run the requested task."""
        if not super().can_run_on(worker_metadata):
            return False

        available_key = self.prefix_with_task_name("available")
        if not worker_metadata.get(available_key, False):
            return False

        executor_available_key = f"executor:{self.backend}:available"
        if not worker_metadata.get(executor_available_key, False):
            return False
        if self.backend != BackendType.UNSHARE:
            if not worker_metadata.get("autopkgtest:available", False):
                return False

        return True

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Download the source artifact."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")
        assert self.dynamic_data

        source_artifact_id = self.dynamic_data.input_source_artifact_id
        artifact = self.debusine.artifact_get(
            source_artifact_id, workspace=self.workspace_name
        )
        if artifact.category != ArtifactCategory.SOURCE_PACKAGE:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"input.source_artifact points to a "
                    f"{artifact.category}, not the expected "
                    f"{ArtifactCategory.SOURCE_PACKAGE}."
                ],
            )
            return False
        for file in artifact.files:
            if file.endswith(".dsc"):
                self._dsc_file = destination / file
                break
        else:
            raise AssertionError("No .dsc file found in source package.")
        self.fetch_artifact(source_artifact_id, destination)

        for artifact_id in self.dynamic_data.input_extra_binary_artifacts_ids:
            artifact = self.debusine.artifact_get(
                artifact_id, workspace=self.workspace_name
            )
            if artifact.category in (
                ArtifactCategory.BINARY_PACKAGE,
                ArtifactCategory.UPLOAD,
            ):
                for file in artifact.files:
                    if file.endswith(".deb"):
                        self._extra_packages.append(destination / file)
            else:
                self.append_to_log_file(
                    "fetch_input.log",
                    [
                        f"input.extra_binary_artifacts includes artifact "
                        f"{artifact_id}, a {artifact.category}, not the "
                        f"expected {ArtifactCategory.BINARY_PACKAGE}, or "
                        f"{ArtifactCategory.UPLOAD}."
                    ],
                )
                return False
            self.fetch_artifact(artifact_id, destination)

        self._prepare_executor()

        self.write_extra_repository_config(
            codename=self._environment_codename(), destination=destination
        )

        return True

    def _environment_codename(self) -> str:
        """Return the codename of the environment we're building in."""
        if self.executor is None:
            raise ValueError(
                "self.executor not set, call _prepare_executor() first"
            )

        return cast(str, self.executor.system_image.data["codename"])

    @override
    def _cmdline(self) -> list[str]:
        """
        Build the sbuild command line.

        Use self.data and self._dsc_file.
        """
        cmd = [
            self.builder,
            "--purge-deps=never",
            # Use a higher-level workflow instead to run both sbuild and
            # lintian, such as debian_pipeline.
            "--no-run-lintian",
        ]
        if "any" in self.data.build_components:
            cmd.append("--arch-any")
        else:
            cmd.append("--no-arch-any")
        if "all" in self.data.build_components:
            cmd.append("--arch-all")
        else:
            cmd.append("--no-arch-all")
        if "source" in self.data.build_components:
            cmd.append("--source")
        else:
            cmd.append("--no-source")

        # TODO: This should probably be host_architecture, if it isn't "all"
        # or "source".
        cmd.append("--arch=" + self.data.build_architecture)
        cmd += self._cmdline_backend()
        cmd += self._cmdline_extra_repository()
        cmd += self._cmdline_binnmu()

        if self.data.build_profiles:
            cmd.append(f"--profiles={','.join(self.data.build_profiles)}")

        if self.data.build_dep_resolver:
            cmd.append(f"--build-dep-resolver={self.data.build_dep_resolver}")
        if self.data.aspcud_criteria:
            cmd.append(f"--aspcud-criteria={self.data.aspcud_criteria}")
        if self.data.resolve_alternatives:
            cmd.append("--resolve-alternatives")

        # BD-Uninstallable
        cmd.append("--bd-uninstallable-explainer=dose3")

        for package in self._extra_packages:
            cmd.append(f"--extra-package={package}")

        cmd.append(str(self._dsc_file))

        return cmd

    def _cmdline_backend(self) -> list[str]:
        """Generate command line arguments for the backend and environment."""
        distribution = self._environment_codename()
        args: list[str] = [f"--dist={distribution}"]

        assert self.executor
        if self.backend == BackendType.UNSHARE:
            args += [
                "--chroot-mode=unshare",
                f"--chroot={self.executor.image_name()}",
                # Remove any dangling resolv.conf symlink (from
                # systemd-resolved installed in the environment, #1071736)
                "--chroot-setup-commands=rm -f /etc/resolv.conf",
            ]
        else:
            virt_server = self.executor.autopkgtest_virt_server()
            args += [
                "--chroot-mode=autopkgtest",
                f"--autopkgtest-virt-server=autopkgtest-virt-{virt_server}",
            ] + [
                f"--autopkgtest-virt-server-opt={opt.replace('%', '%%')}"
                for opt in self.executor.autopkgtest_virt_args()
            ]
        return args

    def _cmdline_extra_repository(self) -> list[str]:
        """Generate the command line arguments for extra_repositories."""
        args: list[str] = []

        # sbuild has no native deb822 support, yet: #1089735
        for path in self.extra_repository_sources:
            inner_path = f"/etc/apt/sources.list.d/{path.name}"
            inner_command = f"cat > {shlex.quote(inner_path)}"
            args.append(
                f"--pre-build-commands=cat {shlex.quote(str(path))} | "
                f"%SBUILD_CHROOT_EXEC sh -c {shlex.quote(inner_command)}"
            )

        # apt < 2.3.10 has no support for keys embedded in Signed-By
        for path in self.extra_repository_keys:
            inner_path = f"/etc/apt/keyrings/{path.name}"
            inner_command = (
                f"mkdir -p /etc/apt/keyrings && cat > {shlex.quote(inner_path)}"
            )
            args.append(
                f"--pre-build-commands=cat {shlex.quote(str(path))} | "
                f"%SBUILD_CHROOT_EXEC sh -c {shlex.quote(inner_command)}"
            )

        return args

    def _cmdline_binnmu(self) -> list[str]:
        """Generate the command line arguments for binnmu."""
        if self.data.binnmu is None:
            return []

        args: list[str] = [
            f"--make-binNMU={self.data.binnmu.changelog}",
            f"--append-to-version={self.data.binnmu.suffix}",
            "--binNMU=0",
        ]
        if self.data.binnmu.timestamp is not None:
            rfc5322date = email.utils.format_datetime(
                self.data.binnmu.timestamp
            )
            args.append(f"--binNMU-timestamp={rfc5322date}")
        maintainer: str | None
        if self.data.binnmu.maintainer is not None:
            maintainer = str(self.data.binnmu.maintainer)
        else:
            assert self.dynamic_data
            maintainer = self.dynamic_data.binnmu_maintainer
        args.append(f"--maintainer={maintainer}")
        return args

    def _build_options(self) -> list[str]:
        return [
            profile
            for profile in self.data.build_profiles or []
            if profile in ("nocheck", "nodoc")
        ]

    @override
    def _cmd_env(self) -> dict[str, str]:
        """Set DEB_BUILD_OPTIONS for build profiles."""
        env = super()._cmd_env()
        if build_options := self._build_options():
            env["DEB_BUILD_OPTIONS"] = " ".join(
                os.environ.get("DEB_BUILD_OPTIONS", "").split() + build_options
            )
        return env

    @override
    def task_result(
        self, returncode: int | None, execute_directory: Path
    ) -> WorkRequestResults:
        """
        Check whether the task succeeded.

        `sbuild` returns 0 for "Status: skipped" (i.e. host architecture not
        supported by source package); we consider that a failure.
        """
        super_result = super().task_result(returncode, execute_directory)
        if super_result != WorkRequestResults.SUCCESS:
            return super_result

        build_log_path = debusine.utils.find_file_suffixes(
            execute_directory, [".build"]
        )
        if build_log_path is None:
            return WorkRequestResults.ERROR

        status = ""
        # Reading the whole build log is inefficient, but shouldn't usually
        # be too bad in practice.
        with open(build_log_path, errors="replace") as f:
            for line in f:
                if line.startswith("Status: "):
                    status = line[len("Status: ") :].rstrip("\n")
        if status == "successful":
            return WorkRequestResults.SUCCESS
        return WorkRequestResults.FAILURE

    @staticmethod
    def _extract_dose3_explanation(
        build_log_path: Path,
    ) -> DoseDistCheck | None:
        """Isolate and parse dose3 output in BD-Uninstallable scenario."""
        output = ''
        with open(build_log_path, errors='replace') as f:
            # find start of dose-debcheck output
            while (line := f.readline()) != '':
                if line == '(I)Dose_applications: Solving...\n':
                    break

            # exit if no/invalid output
            if not (line := f.readline()).startswith('output-version'):
                return None
            output += line

            # grab until next section
            while (line := f.readline()) != '':
                if line.startswith('+----'):
                    break
                output += line

            # convert & validate to Pydantic structure
            # Use yaml.CBaseLoad to work-around malformed yaml
            # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=834059#22
            # https://gitlab.com/irill/dose3/-/issues/18
            parsed_dict = yaml.load(output, yaml.CBaseLoader)
            try:
                return DoseDistCheck.model_validate(parsed_dict)
            except pydantic.ValidationError:
                return None

    def _upload_package_build_log(
        self,
        build_directory: Path,
        source: str,
        version: str,
        architecture: str,
        execution_result: WorkRequestResults,
    ) -> RemoteArtifact | None:
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        build_log_path = debusine.utils.find_file_suffixes(
            build_directory, [".build"]
        )
        if build_log_path is None:
            return None

        explanation = None
        if execution_result != WorkRequestResults.SUCCESS:
            # BD-Uninstallable: look for the dose3 output
            explanation = self._extract_dose3_explanation(build_log_path)

        package_build_log = PackageBuildLog.create(
            source=source,
            version=version,
            architecture=architecture,
            file=build_log_path,
            bd_uninstallable=explanation,
        )

        return self.debusine.upload_artifact(
            package_build_log,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

    def _upload_binary_upload(
        self, build_directory: Path
    ) -> RemoteArtifact | None:
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        changes_path = debusine.utils.find_file_suffixes(
            build_directory, [".changes"]
        )
        if changes_path is None:
            return None

        artifact_binary_upload = Upload.create(
            changes_file=changes_path,
        )
        return self.debusine.upload_artifact(
            artifact_binary_upload,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

    def _upload_signing_input(
        self, build_directory: Path
    ) -> RemoteArtifact | None:
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        changes_path = debusine.utils.find_file_suffixes(
            build_directory, [".changes"]
        )
        if changes_path is None:
            return None

        artifact_signing_input = SigningInputArtifact.create(
            [changes_path], build_directory
        )
        return self.debusine.upload_artifact(
            artifact_signing_input,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

    def _create_binary_package_local_artifacts(
        self,
        build_directory: Path,
        suffixes: list[str],
    ) -> list[BinaryPackage]:
        deb_paths = debusine.utils.find_files_suffixes(
            build_directory, suffixes
        )

        artifacts: list[BinaryPackage] = []
        for deb_path in deb_paths:
            artifacts.append(BinaryPackage.create(file=deb_path))  # noqa: PERF401
        return artifacts

    def _upload_binary_packages(
        self,
        build_directory: Path,
    ) -> list[RemoteArtifact]:
        r"""Upload \*.deb and \*.udeb files."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        host_arch = self._host_architecture

        packages = []

        if "any" in self.data.build_components:
            prefix = "_" + host_arch
            packages.extend(
                self._create_binary_package_local_artifacts(
                    build_directory,
                    [prefix + ".deb", prefix + ".udeb"],
                )
            )

        if "all" in self.data.build_components:
            prefix = "_all"
            packages.extend(
                self._create_binary_package_local_artifacts(
                    build_directory,
                    [prefix + ".deb", prefix + ".udeb"],
                )
            )

        remote_artifacts: list[RemoteArtifact] = []
        for package in packages:
            remote = self.debusine.upload_artifact(
                package,
                workspace=self.workspace_name,
                work_request=self.work_request_id,
            )
            remote_artifacts.append(remote)
            self._uploaded_binary_packages.append((package, remote))

        return remote_artifacts

    @staticmethod
    def _find_matching_dbgsym(directory: Path, binary_name: str) -> Path | None:
        if not binary_name.endswith(".deb"):
            return None

        stem = binary_name[: -len(".deb")]
        if "_" not in stem:
            return None

        package, rest = stem.split("_", 1)
        for ext in (".deb", ".ddeb"):
            candidate = directory / f"{package}-dbgsym_{rest}{ext}"
            if candidate.is_file():
                return candidate
        return None

    @classmethod
    def _extract_build_id_files(
        cls, *, dbgsym_path: Path, destination: Path
    ) -> dict[str, Path]:
        build_id_files: dict[str, Path] = {}
        with debfile.DebFile(dbgsym_path) as pkg:
            tar = pkg.data.tgz()
            for member in tar.getmembers():
                if not member.isreg():
                    continue
                if not member.name.startswith(_BUILD_ID_DEBUG_PREFIX):
                    continue
                if (
                    match := _BUILD_ID_DEBUG_PATH_RE.match(member.name)
                ) is None:
                    raise DebugSymbolPathError(
                        f"{dbgsym_path.name}: malformed .build-id "
                        f"path {member.name!r}; expected "
                        f"./usr/lib/debug/.build-id/<XX>/<YYYY>.debug"
                    )
                build_id = match.group("prefix") + match.group("suffix")
                out_path = destination / f"{build_id}.debug"
                src = tar.extractfile(member)
                if src is None:  # pragma: no cover - defensive
                    continue
                try:
                    with out_path.open("wb") as dst:
                        shutil.copyfileobj(src, dst)
                finally:
                    src.close()
                build_id_files[build_id] = out_path
        return build_id_files

    def _upload_debug_symbols(
        self,
        build_directory: Path,
    ) -> list[tuple[RemoteArtifact, RemoteArtifact]]:
        """
        Upload one ``debian:debug-symbols`` artifact per -dbgsym .deb.

        Iterates over binaries already uploaded by
        :meth:`_upload_binary_packages`.

        :return: ``(debug_remote, parent_binary_remote)`` pairs for
            :meth:`_create_debug_symbols_relations` — debug-symbols artifact
            first, parent binary package second.
        """
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        pairs: list[tuple[RemoteArtifact, RemoteArtifact]] = []

        for binary_local, binary_remote in self._uploaded_binary_packages:
            for binary_name in binary_local.files:
                dbgsym_path = self._find_matching_dbgsym(
                    build_directory, binary_name
                )
                if dbgsym_path is None:
                    continue

                with tempfile.TemporaryDirectory(
                    prefix="debusine-dbgsym-"
                ) as raw_tmp:
                    build_id_files = self._extract_build_id_files(
                        dbgsym_path=dbgsym_path,
                        destination=Path(raw_tmp),
                    )
                    if not build_id_files:
                        continue

                    local_debug = DebugSymbols.create(
                        build_id_files=build_id_files
                    )
                    remote_debug = self.debusine.upload_artifact(
                        local_debug,
                        workspace=self.workspace_name,
                        work_request=self.work_request_id,
                    )

                pairs.append((remote_debug, binary_remote))

        return pairs

    def _create_debug_symbols_relations(
        self,
        debug_symbol_pairs: list[tuple[RemoteArtifact, RemoteArtifact]],
    ) -> None:
        """Create RELATES_TO from each debug-symbols artifact to its binary."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        for debug_remote, binary_remote in debug_symbol_pairs:
            self.debusine.relation_create(
                debug_remote.id,
                binary_remote.id,
                RelationType.RELATES_TO,
            )

    def _create_remote_binary_packages_relations(
        self,
        remote_build_log: RemoteArtifact | None,
        remote_binary_upload: RemoteArtifact | None,
        remote_binary_packages: list[RemoteArtifact],
        remote_signing_input: RemoteArtifact | None,
    ) -> None:
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        for remote_binary_package in remote_binary_packages:
            for source_artifact_id in self._source_artifacts_ids:
                self.debusine.relation_create(
                    remote_binary_package.id,
                    source_artifact_id,
                    RelationType.BUILT_USING,
                )

            if remote_build_log is not None:
                self.debusine.relation_create(
                    remote_build_log.id,
                    remote_binary_package.id,
                    RelationType.RELATES_TO,
                )

            if remote_binary_upload is not None:
                self.debusine.relation_create(
                    remote_binary_upload.id,
                    remote_binary_package.id,
                    RelationType.EXTENDS,
                )
                self.debusine.relation_create(
                    remote_binary_upload.id,
                    remote_binary_package.id,
                    RelationType.RELATES_TO,
                )
                if remote_build_log is not None:
                    self.debusine.relation_create(
                        remote_build_log.id,
                        remote_binary_upload.id,
                        RelationType.RELATES_TO,
                    )

        if (
            remote_binary_upload is not None
            and remote_signing_input is not None
        ):
            self.debusine.relation_create(
                remote_signing_input.id,
                remote_binary_upload.id,
                RelationType.RELATES_TO,
            )

    @override
    def configure_for_execution(
        self,
        download_directory: Path,  # noqa: U100
    ) -> bool:
        """
        Configure Task: set variables needed for the build() step.

        Return True if configuration worked, False, if there was a problem.
        """
        if self._dsc_file is None or not self._dsc_file.exists():
            self.append_to_log_file(
                "configure_for_execution.log",
                ["Input source package not found."],
            )
            return False

        return True

    @override
    def upload_artifacts(
        self, directory: Path, *, execution_result: WorkRequestResults
    ) -> None:
        """Upload artifacts for the task.

        :param directory: directory containing the files that
          will be uploaded.
        :param execution_result: if this is not ``WorkRequestResults.SUCCESS``,
          skip uploading ``.changes`` and ``*.deb``/``*.udeb``.
        """
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        if self._dsc_file is None:
            return
        # ValueError should be impossible here, since it would already have
        # been caught by
        # SbuildValidatorMixin.check_directory_for_consistency_errors.  If
        # it somehow happens anyway, just let it propagate up.
        dsc = read_dsc(self._dsc_file)

        # Upload the .build file (PackageBuildLog)
        remote_build_log = self._upload_package_build_log(
            directory,
            dsc["source"],
            dsc["version"],
            self._host_architecture,
            execution_result,
        )

        if remote_build_log is not None:
            for source_artifact_id in self._source_artifacts_ids:
                self.debusine.relation_create(
                    remote_build_log.id,
                    source_artifact_id,
                    RelationType.RELATES_TO,
                )

        if execution_result == WorkRequestResults.SUCCESS:
            # Upload the *.deb/*.udeb files (BinaryPackages)
            remote_binary_packages = self._upload_binary_packages(directory)

            # Upload debian:debug-symbols artifacts extracted from
            # -dbgsym .debs produced by sbuild
            debug_symbol_pairs = self._upload_debug_symbols(directory)

            # Upload the .changes and the rest of the files
            remote_binary_changes = self._upload_binary_upload(directory)

            # Upload the .changes on its own as signing input
            remote_signing_input = self._upload_signing_input(directory)

            # Create the relations
            self._create_remote_binary_packages_relations(
                remote_build_log,
                remote_binary_changes,
                remote_binary_packages,
                remote_signing_input,
            )
            self._create_debug_symbols_relations(debug_symbol_pairs)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utilities for debusine's dput-ng integration."""

from typing import Any

from dput.core import logger

from debusine.client.config import ConfigHandler, ServerInfo
from debusine.client.debusine import Debusine


def get_debusine_client_config(
    fqdn: str, scope_name: str | None = None
) -> ServerInfo:
    """
    Get debusine client configuration for a given FQDN.

    This is a useful hook for testing.
    """
    configuration = ConfigHandler(server_name=f"{fqdn}/{scope_name}")
    return configuration.server_configuration()


def make_debusine_client(profile: dict[str, Any]) -> Debusine:
    """Make a suitable debusine client."""
    fqdn = profile["fqdn"]
    scope = profile["debusine_scope"]

    config = get_debusine_client_config(fqdn, scope_name=scope)
    return Debusine(
        base_api_url=config.api_url,
        api_token=config.api_token,
        scope=scope,
        logger=logger,
    )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Rename ``host_architecture`` to ``build_architecture`` in more places."""

from django.db import migrations

from debusine.artifacts.models import TaskTypes
from debusine.db.migrations._utils import make_data_field_renamer

# Same as in 0029_rename_build_architecture.
affected_tasks = {
    "autopkgtest",
    "lintian",
    "blhc",
    "debdiff",
    "piuparts",
    "sbuild",
    "extractforsigning",
    "assemblesignedsource",
    "makesourcepackageupload",
}


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0029_rename_build_architecture"),
    ]

    operations = [
        make_data_field_renamer(
            "WorkRequest",
            "dynamic_task_data",
            {"task_type": TaskTypes.WORKER, "task_name__in": affected_tasks},
            [("host_architecture", "build_architecture")],
        ),
        make_data_field_renamer(
            "WorkRequest",
            "configured_task_data",
            {"task_type": TaskTypes.WORKER, "task_name__in": affected_tasks},
            [("host_architecture", "build_architecture")],
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Exception based HTTP error handling."""

from collections.abc import Callable
from functools import wraps
from typing import Any

from django.http import HttpRequest
from django.http.response import HttpResponseBase
from django.template.response import TemplateResponse
from rest_framework import status


class HttpError(Exception):
    """HTTP error to be rendered as a response."""

    def __init__(self, code: int, error: str) -> None:
        """Initialize with http code and error message."""
        super().__init__(error)
        self.code = code


class HttpError400(HttpError):
    """Raise a Bad Request HTTP error."""

    def __init__(self, error: str) -> None:
        """Force code to be 400."""
        super().__init__(code=status.HTTP_400_BAD_REQUEST, error=error)


def catch_http_errors(
    view_func: Callable[..., HttpResponseBase],
) -> Callable[..., HttpResponseBase]:
    """Wrap a view function to render raised HttpError exceptions."""
    # We mostly use this of HTTP 400, for which there is BadRequest. However,
    # django.views.defaults.bad_request does not pass information from the
    # exception to the template context, to prevent disclosing sensitive
    # information.
    #
    # We can use this decorator to have an alternative exception-based error
    # reporting for views, using a custom HttpError exception that is intended
    # to be shown to the user

    @wraps(view_func)
    def wrapped(
        request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        try:
            return view_func(request, *args, **kwargs)
        except HttpError as e:
            context = {"error": e.args[0]}
            return TemplateResponse(
                request, f"{e.code}.html", context, status=e.code
            )

    return wrapped



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for server-side task models."""

from unittest import TestCase

from debusine.server.tasks.models import APTMirrorData


class APTMirrorDataTests(TestCase):
    """Tests for APTMirror task data."""

    def test_flat_repository(self) -> None:
        """A flat repository must not have components."""
        APTMirrorData(
            collection="example",
            url="https://deb.example.org/",
            suite="./",
            architectures=["amd64"],
        )
        with self.assertRaisesRegex(
            ValueError,
            r'Flat repositories \(where suite ends with "/"\) must not have '
            r'components',
        ):
            APTMirrorData(
                collection="example",
                url="https://deb.example.org/",
                suite="./",
                components=["main"],
                architectures=["amd64"],
            )

    def test_non_flat_repository(self) -> None:
        """A non-flat repository must have components."""
        APTMirrorData(
            collection="debian/bookworm",
            url="https://deb.debian.org/debian",
            suite="bookworm",
            components=["main"],
            architectures=["amd64"],
        )
        with self.assertRaisesRegex(
            ValueError,
            r'Non-flat repositories \(where suite does not end with "/"\) must '
            r'have components',
        ):
            APTMirrorData(
                collection="debian/bookworm",
                url="https://deb.debian.org/debian",
                suite="bookworm",
                architectures=["amd64"],
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Empty Workflow, used for testing."""

from typing import override

from debusine.server.workflows.base import Workflow
from debusine.server.workflows.models import BaseWorkflowData


class NoopWorkflow(Workflow[BaseWorkflowData]):
    """Workflow that does nothing."""

    TASK_NAME = "noop"

    @override
    def populate(self) -> None:
        """Do nothing."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URLs for the server application - API."""

from django.urls import URLPattern, URLResolver, include, path

from debusine.server.views.artifacts import (
    ArtifactRelationsView,
    ArtifactView,
    UploadFileView,
    WorkspaceArtifactView,
)
from debusine.server.views.assets import AssetPermissionCheckView, AssetView
from debusine.server.views.base import SupersededAPIView, Whoami
from debusine.server.views.collection_relations import (
    CollectionRelationListView,
    CollectionRelationSetView,
)
from debusine.server.views.collections import (
    CollectionTokenGrantViewSet,
    CollectionViewSet,
    TaskConfigurationCollectionView,
)
from debusine.server.views.enroll import EnrollView
from debusine.server.views.lookups import LookupMultipleView, LookupSingleView
from debusine.server.views.open_metrics import (
    OpenMetricsEmitView,
    OpenMetricsView,
)
from debusine.server.views.routers import APIRouter
from debusine.server.views.service_status import ServiceStatusView
from debusine.server.views.work_requests import (
    WorkRequestAbortView,
    WorkRequestConfirmView,
    WorkRequestExternalDebsignView,
    WorkRequestRetryView,
    WorkRequestUnblockView,
    WorkRequestView,
)
from debusine.server.views.workers import (
    GetNextWorkRequestView,
    RegisterView,
    UpdateWorkRequestAsCompletedView,
    UpdateWorkerDynamicMetadataView,
)
from debusine.server.views.workflows import (
    WorkflowTemplateViewSet,
    WorkflowView,
)
from debusine.server.views.workspaces import WorkspaceInheritanceView

app_name = 'server'

urlpatterns_1_0: list[URLPattern | URLResolver] = [
    path(
        'whoami/',
        Whoami.as_view(),
        name='whoami',
    ),
    path(
        'work-request/get-next-for-worker/',
        GetNextWorkRequestView.as_view(),
        name='work-request-get-next',
    ),
    path(
        'work-request/',
        WorkRequestView.as_view(),
        name='work-requests',
    ),
    path(
        'work-request/<int:work_request_id>/',
        WorkRequestView.as_view(),
        name='work-request-detail',
    ),
    path(
        'work-request/<int:work_request_id>/completed/',
        UpdateWorkRequestAsCompletedView.as_view(),
        name='work-request-completed',
    ),
    path(
        'work-request/<int:work_request_id>/retry/',
        WorkRequestRetryView.as_view(),
        name='work-requests-retry',
    ),
    path(
        'work-request/<int:work_request_id>/abort/',
        WorkRequestAbortView.as_view(),
        name='work-request-abort',
    ),
    path(
        'work-request/<int:work_request_id>/unblock/',
        WorkRequestUnblockView.as_view(),
        name='work-request-unblock',
    ),
    path(
        'work-request/<int:work_request_id>/external-debsign/',
        WorkRequestExternalDebsignView.as_view(),
        name='work-request-external-debsign',
    ),
    path(
        "work-request/<int:work_request_id>/confirm/",
        WorkRequestConfirmView.as_view(),
        name="work-request-confirm",
    ),
    path(
        "artifact/",
        ArtifactView.as_view(),
        name="artifact-create",
    ),
    path(
        "artifact/<str:workspace>/<int:artifact_id>/",
        WorkspaceArtifactView.as_view(),
        name="workspace-artifact",
    ),
    path(
        "artifact/<int:artifact_id>/",
        ArtifactView.as_view(),
        name="artifact",
    ),
    path(
        "artifact/<int:artifact_id>/files/<path:file_path>/",
        UploadFileView.as_view(),
        name="upload-file",
    ),
    path(
        "asset/",
        AssetView.as_view(),
        name="asset",
    ),
    path(
        "asset/<str:asset_category>/<str:asset_slug>/<str:permission_name>/",
        AssetPermissionCheckView.as_view(),
        name="asset-permission-check",
    ),
    path("lookup/single/", LookupSingleView.as_view(), name="lookup-single"),
    path(
        "lookup/multiple/",
        LookupMultipleView.as_view(),
        name="lookup-multiple",
    ),
    path('worker/register/', RegisterView.as_view(), name='register'),
    path(
        'worker/dynamic-metadata/',
        UpdateWorkerDynamicMetadataView.as_view(),
        name='worker-dynamic-metadata',
    ),
    path(
        "artifact-relation/",
        ArtifactRelationsView.as_view(),
        name="artifact-relation-list",
    ),
    path(
        "artifact-relation/<int:pk>/",
        ArtifactRelationsView.as_view(),
        name="artifact-relation-detail",
    ),
    path("workflow/", WorkflowView.as_view(), name="workflows"),
    path(
        "service-status/",
        ServiceStatusView.as_view(),
        name="service-status",
    ),
    path(
        "task-configuration/<str:workspace>/<str:name>/",
        TaskConfigurationCollectionView.as_view(),
        name="task-configuration-collection",
    ),
    path(
        "workspace/<str:workspace>/inheritance/",
        WorkspaceInheritanceView.as_view(),
        name="workspace-inheritance",
    ),
    path(
        "open-metrics/",
        OpenMetricsView.as_view(),
        name="open-metrics",
    ),
    path(
        "open-metrics/emit/",
        OpenMetricsEmitView.as_view(),
        name="open-metrics-emit",
    ),
    # Cannot use it as a viewset: it's a set of collection relations
    path(
        "collection-relation/<str:workspace>/<str:collection>/<str:relation_type>/",
        CollectionRelationSetView.as_view(),
        name="collection-relation-update",
    ),
    # Deprecated: use the workflow-template viewset instead
    path(
        "workflow-template/",
        SupersededAPIView.as_view(),
        {
            "new_endpoint": "/1.0/workflow-template/<workspace>/",
            "client_version": "0.13.0",
        },
        name="workflow-templates",
    ),
]

router = APIRouter()
router.include_root_view = False
router.register(
    "collection/(?P<workspace>[^/]+)",
    CollectionViewSet,
    basename="collection",
)
router.register(
    "collection/(?P<workspace>[^/]+)/(?P<ccat>[^/]+)/(?P<cname>[^/]+)/token",
    CollectionTokenGrantViewSet,
    basename="collection-token-grant",
)
router.register(
    "collection-relation/(?P<workspace>[^/]+)",
    CollectionRelationListView,
    basename="collection-relation",
)
router.register(
    "workflow-template/(?P<workspace>[^/]+)",
    WorkflowTemplateViewSet,
    basename="workflow-template",
)
urlpatterns_1_0 += router.urls

urlpatterns = [
    path(
        'auth/',
        include('rest_framework.urls', namespace='rest_framework'),
    ),
    path("enroll/", EnrollView.as_view(), name="enroll"),
    path("1.0/", include(urlpatterns_1_0)),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the workflow templates views."""

from typing import ClassVar, override

import yaml
from rest_framework import status

from debusine.db.models import WorkflowTemplate
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.tests.utils import ViewTestMixin


class WorkflowTemplateDetailViewTests(ViewTestMixin, TestCase):
    """Tests for WorkflowTemplateDetailView class."""

    template: ClassVar[WorkflowTemplate]
    scenario = scenarios.DefaultContext(set_current=True)

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.template = cls.playground.create_workflow_template(
            "name-1",
            "noop",
            workspace=cls.scenario.workspace,
            static_parameters={"succeeds": "True"},
            runtime_parameters={},
        )

    def test_permissions(self) -> None:
        """Test that the right permissions are enforced."""
        self.assertSetsCurrentWorkspace(
            self.scenario.workspace,
            self.template.get_absolute_url(),
        )
        self.assertEnforcesPermission(
            self.template.can_display,
            self.template.get_absolute_url(),
            "get_context_data",
        )

    def test_detail_view(self) -> None:
        """Test detail view."""
        response = self.client.get(self.template.get_absolute_url())
        tree = self.assertResponseHTML(response)

        h1 = self.assertHasElement(tree, "//h1[1]")
        self.assertEqual(h1, f"Workflow template {self.template.name}")

        table = self.assertHasElement(
            tree, '//table[@id="workflow-information"]'
        )

        name = self.assertHasElement(table, ".//tr[1]/td[1]")
        self.assertTextContentEqual(name, self.template.name)

        task_name = self.assertHasElement(table, ".//tr[2]/td[1]")
        self.assertTextContentEqual(task_name, self.template.task_name)

        self.assertEqual(task_name.a, self.template.task_name)
        self.assertEqual(
            task_name.a.attrib["href"],
            f"https://freexian-team.pages.debian.net/debusine/"
            f"reference/workflows/specs/{self.template.task_name}.html",
        )

        priority = self.assertHasElement(table, ".//tr[3]/td[1]")
        self.assertTextContentEqual(priority, str(self.template.priority))

        static_parameters = self.assertHasElement(table, ".//tr[5]/td[1]")
        self.assertEqual(
            yaml.safe_load(self.get_node_text_normalized(static_parameters)),
            self.template.static_parameters,
        )

        runtime_parameters = self.assertHasElement(table, ".//tr[7]/td[1]")
        self.assertEqual(
            yaml.safe_load(self.get_node_text_normalized(runtime_parameters)),
            self.template.runtime_parameters,
        )

    def test_namespaced_by_workspace(self) -> None:
        # Create another workflow template with the same name, in a
        # neighbouring workspace
        self.playground.create_workflow_template(
            self.template.name,
            "noop",
            workspace=self.playground.create_workspace(name="another"),
            static_parameters={"succeeds": "True"},
        )
        response = self.client.get(self.template.get_absolute_url())

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_does_not_exist(self) -> None:
        """A nonexistent workflow template returns 404."""
        response = self.client.get(
            f"/{self.scenario.workspace}/workflow-template/nonexistent/"
        )

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"],
            "No WorkflowTemplate matches the given query.",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Worker client to connect to Debusine."""

import sys
from typing import cast

from debusine.worker._worker import Worker

# Import the documentation from where the code lives
__doc__ = cast(str, sys.modules[Worker.__module__].__doc__)

__all__ = [
    'Worker',
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Lookups of items in collections."""

import re
from collections.abc import Sequence
from typing import Any, Literal, assert_never, overload, override

from django.db.models import Q
from django.db.models.fields.json import KT

from debusine.artifacts.models import BareDataCategory, CollectionCategory
from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    WorkRequest,
    Workspace,
)
from debusine.db.models.lookups import (
    LookupResult,
    LookupResultArtifact,
    LookupResultBare,
    LookupResultCollection,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.tasks.models import (
    LookupDict,
    LookupMultiple,
    LookupSingle,
    build_lookup_string,
    parse_lookup_string,
)

LOOKUP_RESULT_TYPE_MAP: dict[CollectionItem.Types, LookupResultType] = {
    CollectionItem.Types.ARTIFACT: LookupResultType.ARTIFACT,
    CollectionItem.Types.BARE: LookupResultType.BARE,
    CollectionItem.Types.COLLECTION: LookupResultType.COLLECTION,
}


class NoSuchItemError(KeyError):
    """
    A variant of KeyError with more suitable stringification.

    Single lookups raise KeyError if the lookup does not resolve to an item,
    and LookupError if the lookup is invalid in some way or returns the
    wrong item type.  However, KeyError has a ``__str__`` implementation
    specialized for the common case of missing items from dictionaries and
    the like.  Since our exceptions include an explanatory string instead,
    bypass this in favour of the default stringification for exceptions.
    """

    @override
    def __str__(self) -> str:
        return LookupError.__str__(self)


def _lookup_result_expected(
    result: LookupResult, expect_type: LookupChildType
) -> bool:
    """Check whether a lookup result is of the expected type."""
    if expect_type != LookupChildType.ANY:
        expected_result_types = {
            LookupChildType.BARE: [LookupResultType.BARE],
            LookupChildType.ARTIFACT: [LookupResultType.ARTIFACT],
            LookupChildType.ARTIFACT_OR_PROMISE: [
                LookupResultType.ARTIFACT,
                LookupResultType.BARE,
            ],
            LookupChildType.COLLECTION: [LookupResultType.COLLECTION],
        }[expect_type]
        if result.result_type not in expected_result_types:
            return False
    return True


def _normalize_lookup_segments(
    lookup: LookupSingle,
    default_category: CollectionCategory | None = None,
    expect_type: LookupChildType = LookupChildType.ANY,
) -> tuple[list[str], dict[str, Any] | None]:
    """Normalize a single lookup into a list of segments."""
    if isinstance(lookup, int):
        if expect_type in (
            LookupChildType.ARTIFACT,
            LookupChildType.ARTIFACT_OR_PROMISE,
        ):
            lookup = f"{lookup}@artifacts"
        elif expect_type == LookupChildType.COLLECTION:
            lookup = f"{lookup}@collections"
        else:
            raise LookupError(
                "Integer lookups only work in contexts that expect an "
                "artifact or a collection"
            )

    if not lookup:
        raise LookupError("Empty lookup")

    try:
        segments, options = parse_lookup_string(lookup)
    except ValueError:
        raise LookupError(
            f"Invalid syntax in lookup options part of {lookup!r}"
        )

    first_segment = segments[0]
    if "@" not in first_segment:
        if default_category is not None:
            first_segment = f"{first_segment}@{default_category}"
        else:
            raise LookupError(
                f"{first_segment!r} does not specify a category and the "
                f"context does not supply a default"
            )

    return ([first_segment, *segments[1:]], options)


def _lookup_single_first(
    lookup: str,
    workspace: Workspace,
    *,
    original_lookup: LookupSingle,
    pc: PermissionContext,
    workflow_root: WorkRequest | None = None,
) -> LookupResult:
    """
    Resolve the first segment of a string lookup.

    This is handled differently from the other segments, because the lookup
    doesn't happen in the context of a collection.
    """
    try:
        if m := re.match(r"^([0-9]+)@artifacts$", lookup):
            # TODO: filter using the current scope?
            # TODO: support user is None to mean the user in the current
            #       context?
            workspaces = Workspace.objects.can_display(pc)
            workspace_filter = Q(workspace__in=workspaces)
            visible_artifacts = Artifact.objects.filter(workspace_filter)
            return LookupResultArtifact(
                lookup=original_lookup,
                artifact=visible_artifacts.get(id=int(m.group(1))),
            )
        elif m := re.match(r"^([0-9]+)@collections$", lookup):
            # TODO: filter using the current scope?
            # TODO: support user is None to mean the user in the current
            #       context?
            workspaces = Workspace.objects.can_display(pc)
            workspace_filter = Q(workspace__in=workspaces)
            visible_collections = Collection.objects.filter(workspace_filter)
            return LookupResultCollection(
                lookup=original_lookup,
                collection=visible_collections.get(id=int(m.group(1))),
            )
        else:
            if lookup == "internal@collections":
                if workflow_root is None:
                    raise LookupError(
                        "internal@collections is only valid in the context of "
                        "a workflow"
                    )
                name = f"workflow-{workflow_root.id}"
                category = str(CollectionCategory.WORKFLOW_INTERNAL)
            else:
                # Ensured by lookup_single.
                assert "@" in lookup
                name, category = lookup.rsplit("@", 1)
            return LookupResultCollection(
                lookup=original_lookup,
                collection=workspace.get_collection(
                    pc=pc, category=category, name=name
                ),
            )
    except (Artifact.DoesNotExist, Collection.DoesNotExist):
        raise NoSuchItemError(f"{lookup!r} does not exist or is hidden")


def _process_lookup_options(
    options: dict[str, Any], workspace: Workspace, pc: PermissionContext
) -> None:
    for name, value in options.items():
        match name:
            case "scope":
                if "workspace" not in options:
                    raise LookupError(
                        "The 'scope' lookup option requires the 'workspace' one"
                    )
            case "workspace":
                scope_name = options.get("scope", workspace.scope.name)
                try:
                    requested_workspace = Workspace.objects.can_display(pc).get(
                        name=value, scope__name=scope_name
                    )
                    options["workspace"] = requested_workspace
                except Workspace.DoesNotExist:
                    raise LookupError(
                        f"Invalid or hidden workspace in lookup options: "
                        f"scope={scope_name!r} workspace={value!r}"
                    )
            case _:
                raise LookupError(f"Unknown lookup option: {name!r}")


@overload
def lookup_single(
    lookup: LookupSingle,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: Literal[LookupChildType.BARE],
    precache_result: LookupResultBare | None = None,
) -> LookupResultBare: ...


@overload
def lookup_single(
    lookup: LookupSingle,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: Literal[LookupChildType.ARTIFACT],
    precache_result: LookupResultArtifact | None = None,
) -> LookupResultArtifact: ...


@overload
def lookup_single(
    lookup: LookupSingle,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: Literal[LookupChildType.COLLECTION],
    precache_result: LookupResultCollection | None = None,
) -> LookupResultCollection: ...


@overload
def lookup_single(
    lookup: LookupSingle,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: LookupChildType = LookupChildType.ANY,
    precache_result: LookupResult | None = None,
) -> LookupResult: ...


def lookup_single(
    lookup: LookupSingle,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: LookupChildType = LookupChildType.ANY,
    precache_result: LookupResult | None = None,
) -> LookupResult:
    """
    Look up a collection item using a single lookup syntax.

    See :ref:`lookup-single`.

    :param precache_result: Rather than doing a normal lookup, insert this
      result into the lookup cache and return it.  This is useful to improve
      workflow performance in some situations, such as workflows that do
      bulk database lookups and then reconstruct lookups to pass down to
      their children.
    :raises KeyError: if the lookup does not resolve to an item.
    :raises LookupError: if the lookup is invalid in some way, or does not
      resolve to the expected collection item type.
    """
    # TODO: make PermissionContext hashable and use that in the cache key
    cache = context.lookup_cache.single
    cache_key = (
        lookup,
        workspace,
        pc.user,
        default_category,
        workflow_root,
        expect_type,
    )
    if precache_result is not None:
        cache[cache_key] = precache_result
        return precache_result
    if (cached := cache.get(cache_key)) is not None:
        return cached

    segments, options = _normalize_lookup_segments(
        lookup, default_category=default_category, expect_type=expect_type
    )

    if options:
        _process_lookup_options(options, workspace=workspace, pc=pc)
        workspace = options.get("workspace", workspace)

    # Look up the first segment as a collection by name and category.
    result = _lookup_single_first(
        segments[0],
        workspace,
        original_lookup=lookup,
        pc=pc,
        workflow_root=workflow_root,
    )
    container: Collection | None = None

    # Resolve each subsequent segment by calling `lookup`.
    for i, segment in enumerate(segments[1:], start=1):
        container_name = build_lookup_string(*segments[:i])
        if result.result_type != LookupResultType.COLLECTION:
            raise LookupError(
                f"{container_name!r} is of type"
                f" {result.result_type.name.lower()!r}"
                " instead of expected 'collection'"
            )
        container = result.collection
        assert container is not None
        if ":" not in segment:
            segment = f"name:{segment}"
        item = container.manager.lookup(segment)
        if item is None:
            raise NoSuchItemError(f"{container_name!r} has no item {segment!r}")
        result = LookupResult.create(
            result_type=LOOKUP_RESULT_TYPE_MAP[
                CollectionItem.Types(item.child_type)
            ],
            lookup=lookup,
            parent_collection_lookup=container_name,
            collection_item=item,
            artifact=item.artifact,
            collection=item.collection,
        )

    if not _lookup_result_expected(result, expect_type):
        raise LookupError(
            f"{lookup!r} is of type {result.result_type.name.lower()!r}"
            f" instead of expected {expect_type.name.lower()!r}"
        )

    cache[cache_key] = result
    return result


def _lookup_dict(  # noqa: C901
    lookup: LookupDict,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: LookupChildType = LookupChildType.ANY,
) -> Sequence[LookupResult]:
    """
    Look up collection items using the dictionary lookup syntax.

    Results are returned in a predictable order (currently ordered by
    collection item ID, although this function only guarantees that the
    order will be stable between successive calls).

    See :ref:`lookup-dict`.
    """
    # TODO: use PermissionContext in the cache key instead of user
    cache = context.lookup_cache.dictionary
    cache_key = (
        lookup,
        workspace,
        pc.user,
        default_category,
        workflow_root,
        expect_type,
    )
    if (cached := cache.get(cache_key)) is not None:
        return cached

    parent_collection_segments, options = _normalize_lookup_segments(
        lookup.collection,
        default_category=default_category,
        expect_type=LookupChildType.COLLECTION,
    )

    parent_collection_lookup = build_lookup_string(
        *parent_collection_segments, options=options
    )

    # Find the containing collection.
    collection = lookup_single(
        lookup.collection,
        workspace,
        pc=pc,
        default_category=default_category,
        workflow_root=workflow_root,
        expect_type=LookupChildType.COLLECTION,
    ).collection
    objects = collection.child_items.active()

    # Prepare query conditions.  We don't need to check the workspace here;
    # it's good enough if the item is in a collection we can see.
    match lookup.child_type:
        case LookupChildType.BARE:
            objects = objects.filter(child_type=CollectionItem.Types.BARE)
        case LookupChildType.ARTIFACT:
            objects = objects.filter(child_type=CollectionItem.Types.ARTIFACT)
        case LookupChildType.ARTIFACT_OR_PROMISE:
            objects = objects.filter(
                Q(child_type=CollectionItem.Types.ARTIFACT)
                | Q(
                    child_type=CollectionItem.Types.BARE,
                    category=BareDataCategory.PROMISE,
                )
            )
        case LookupChildType.COLLECTION:
            objects = objects.filter(child_type=CollectionItem.Types.COLLECTION)
        case LookupChildType.ANY:
            pass
        case _ as unreachable:
            assert_never(unreachable)
    if lookup.category is not None:
        objects = objects.filter(category=lookup.category)
    if lookup.name_matcher is not None:
        objects = objects.filter(
            **{f"name__{lookup.name_matcher.kind}": lookup.name_matcher.value}
        )
    for key, matcher in lookup.data_matchers:
        annotation = f"data_text_{key}"
        objects = objects.annotate(**{annotation: KT(f"data__{key}")}).filter(
            **{f"{annotation}__{matcher.kind}": matcher.value}
        )
    for key, value in lookup.lookup_filters:
        objects = objects.filter(
            collection.manager.lookup_filter(
                key,
                value,
                workspace=workspace,
                pc=pc,
                workflow_root=workflow_root,
            )
        )
    objects = objects.select_related("artifact", "collection")

    # Execute the query.
    results = [
        LookupResult.create(
            result_type=LOOKUP_RESULT_TYPE_MAP[
                CollectionItem.Types(item.child_type)
            ],
            lookup=lookup,
            parent_collection_lookup=parent_collection_lookup,
            collection_item=item,
            artifact=item.artifact,
            collection=item.collection,
        )
        for item in objects.order_by("id")
    ]
    if lookup.error_on_empty and not results:
        raise LookupError(f"No results found from {lookup.export()}")
    for result in results:
        if not _lookup_result_expected(result, expect_type):
            raise LookupError(
                f"Result from {lookup.export()} is of type"
                f" {result.result_type.name.lower()!r} instead of expected"
                f" {expect_type.name.lower()!r}"
            )

    frozen_results = tuple(results)
    cache[cache_key] = frozen_results
    return frozen_results


@overload
def lookup_multiple(
    lookup: LookupMultiple,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: Literal[LookupChildType.ARTIFACT],
) -> Sequence[LookupResultArtifact]: ...


@overload
def lookup_multiple(
    lookup: LookupMultiple,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: Literal[LookupChildType.COLLECTION],
) -> Sequence[LookupResultCollection]: ...


@overload
def lookup_multiple(
    lookup: LookupMultiple,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: LookupChildType = LookupChildType.ANY,
) -> Sequence[LookupResult]: ...


def lookup_multiple(
    lookup: LookupMultiple,
    workspace: Workspace,
    *,
    pc: PermissionContext,
    default_category: CollectionCategory | None = None,
    workflow_root: WorkRequest | None = None,
    expect_type: LookupChildType = LookupChildType.ANY,
) -> Sequence[LookupResult]:
    """
    Look up collection items using a multiple lookup syntax.

    See :ref:`lookup-multiple`.

    :raises KeyError: if any of the lookups does not resolve to an item.
    :raises LookupError: if any of the lookups is invalid in some way, or
      does not resolve to the expected collection item type.
    """
    results: list[LookupResult] = []
    for alternative in lookup:
        match alternative:
            case int() | str():
                results.append(
                    lookup_single(
                        alternative,
                        workspace,
                        pc=pc,
                        default_category=default_category,
                        workflow_root=workflow_root,
                        expect_type=expect_type,
                    )
                )
            case LookupDict():
                results.extend(
                    _lookup_dict(
                        alternative,
                        workspace,
                        pc=pc,
                        default_category=default_category,
                        workflow_root=workflow_root,
                        expect_type=expect_type,
                    )
                )
            case _ as unreachable:
                assert_never(unreachable)
    return tuple(results)


def artifact_ids(results: Sequence[LookupResult]) -> list[int]:
    """Return the artifact ids."""
    return [
        result.artifact.id for result in results if result.artifact is not None
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to list workers."""

import warnings
from typing import Any, NoReturn, override

from django.core.management import CommandParser

from debusine.server.management.commands import worker


class Command(worker.Command):
    """Command to list the workers."""

    help = "List all the workers"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            '--yaml', action="store_true", help="Machine readable YAML output"
        )

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Forward to `debusine-admin worker list`."""
        warnings.warn(
            "The `debusine-admin list_workers` command has been deprecated in "
            "favour of `debusine-admin worker list`",
            DeprecationWarning,
        )
        super().handle_list(*args, **options)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Task to create a child workspace.

See docs/reference/devel-blueprints/experiment-workspaces.rst
"""

import datetime as dt
from typing import override

from debusine.artifacts.models import (
    SINGLETON_COLLECTION_CATEGORIES,
    WorkRequestResults,
)
from debusine.db.context import context
from debusine.db.models import Collection, Group, WorkflowTemplate, Workspace
from debusine.server.tasks import BaseServerTask
from debusine.server.tasks.models import CreateChildWorkspaceData


class CannotCreate(Exception):
    """Creation of the new workspace failed."""


class CreateChildWorkspace(BaseServerTask[CreateChildWorkspaceData]):
    """Task that creates a workspace."""

    TASK_VERSION = 1
    TASK_NAME = "create_child_workspace"

    def get_child_workspace_name(self) -> str:
        """Build the name of the child workspace."""
        return f"{self.data.prefix}-{self.data.suffix}"

    @override
    def _execute(self) -> WorkRequestResults:
        """Execute the task."""
        child_workspace_name = self.get_child_workspace_name()
        base_workspace = self.work_request.workspace

        # We need to override permissions later, so we need to check here
        # that the user has the permissions needed to run this server task
        self.enforce(base_workspace.can_create_child_workspace)

        # Lookup workflow templates to copy
        workflow_templates: list[WorkflowTemplate] = []
        for template_name in self.data.workflow_template_names:
            try:
                workflow_templates.append(
                    WorkflowTemplate.objects.get(
                        workspace=base_workspace, name=template_name
                    )
                )
            except WorkflowTemplate.DoesNotExist:
                raise CannotCreate(
                    f"workflow template {base_workspace.name}/{template_name}"
                    " does not exist"
                )

        # Create the new workspace
        with context.disable_permission_checks():
            workspace = Workspace.objects.create(
                scope=base_workspace.scope,
                name=child_workspace_name,
                expiration_delay=(
                    None
                    if self.data.expiration_delay is None
                    else dt.timedelta(days=self.data.expiration_delay)
                ),
                public=self.data.public,
            )

        # Set the new workspace as inheriting an existing workspace
        workspace.set_inheritance([base_workspace])

        # Create the usual singleton collections.  (This could be made
        # optional if somebody has a use case for creating child workspaces
        # without them.)
        for category in SINGLETON_COLLECTION_CATEGORIES:
            Collection.objects.get_or_create_singleton(category, workspace)

        # Lookup or create the owner group
        if self.data.owner_group is None:
            owner_group = Group.objects.create_ephemeral(
                workspace,
                name="Owners",
                owner=self.work_request.created_by,
            )
        else:
            try:
                owner_group = Group.objects.get(
                    name=self.data.owner_group, scope=base_workspace.scope
                )
            except Group.DoesNotExist:
                raise CannotCreate(
                    f"owner group {self.data.owner_group} not found"
                )

        # Set the new group as OWNER of the new workspace
        owner_group.assign_role(workspace, Workspace.Roles.OWNER)

        # Copy WorkflowTemplates to the new workspace
        for workflow_template in workflow_templates:
            # See https://docs.djangoproject.com/en/5.1/topics/db/queries/#copying-model-instances  # noqa: E501
            workflow_template.pk = None
            workflow_template._state.adding = True
            workflow_template.workspace = workspace
            workflow_template.save()

        return WorkRequestResults.SUCCESS

    @override
    def get_task_configuration_subject_context(
        self,
    ) -> tuple[str | None, str | None]:
        # Compute dynamic task to extract subject and
        # configuration_context
        # TODO: these values can be computed in a separate methods if the rest
        # TODO: of compute_dynamic_data becomes too expensive to run twice
        early_dynamic_task_data = self.compute_dynamic_data()
        return (
            early_dynamic_task_data.subject,
            early_dynamic_task_data.configuration_context,
        )

    @override
    def get_label(self) -> str:
        """Return the task label."""
        return f"create child workspace {self.get_child_workspace_name()!r}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for utils.py."""

from debian import deb822

from debusine.artifacts import artifact_utils
from debusine.test import TestCase


class ArtifactUtilsTests(TestCase):
    """Tests for functions in artifact_utils.py."""

    def test_files_in_meta_file_match_files_return_None(self) -> None:
        """Files in files match files in .dsc: return None."""
        file_1 = self.create_temporary_file()

        dsc_file = self.create_temporary_file(suffix=".dsc")

        self.write_dsc_file(dsc_file, files=[file_1])

        files = {
            file_1.name: file_1,
            dsc_file.name: dsc_file,
        }

        # Does not raise an exception.
        artifact_utils.files_in_meta_file_match_files(".dsc", deb822.Dsc, files)

    def test_files_in_meta_file_extra_file_in_files_raise_value_error(
        self,
    ) -> None:
        """Files in files contain extra file (not listed in .dsc)."""
        file_1 = self.create_temporary_file(prefix="b")
        file_2 = self.create_temporary_file(prefix="c")

        dsc_file = self.create_temporary_file(prefix="a", suffix=".dsc")

        self.write_dsc_file(dsc_file, files=[file_2])

        files = {
            dsc_file.name: dsc_file,
            file_1.name: file_1,
        }

        files_str = rf"'{dsc_file.name}', '{file_1.name}'"
        files_listed_in_dsc = rf"'{file_2.name}'"

        expected_msg = (
            rf"^Files in the package and listed in the \.dsc must match. "
            rf"Files: \[{files_str}\] "
            rf"Listed in \.dsc: \[{files_listed_in_dsc}\]$"
        )
        with self.assertRaisesRegex(ValueError, expected_msg):
            artifact_utils.files_in_meta_file_match_files(
                ".dsc", deb822.Dsc, files
            )

    def test_files_in_meta_file_missing_file_in_files_raise_value_error(
        self,
    ) -> None:
        """Files in file misses a file listed in .dsc."""
        file_1 = self.create_temporary_file()

        dsc_file = self.create_temporary_file(suffix=".changes")

        self.write_changes_file(dsc_file, files=[file_1])

        files = {
            dsc_file.name: dsc_file,
        }

        files_str = rf"'{dsc_file.name}'"
        files_listed_in_dsc = rf"'{file_1.name}'"

        expected_msg = (
            rf"^Files in the package and listed in the \.changes must match. "
            rf"Files: \[{files_str}\] "
            rf"Listed in \.changes: \[{files_listed_in_dsc}\]$"
        )

        with self.assertRaisesRegex(ValueError, expected_msg):
            artifact_utils.files_in_meta_file_match_files(
                ".changes", deb822.Changes, files
            )

    def test_file_contains_file_in_changes_make_loop_complete(self) -> None:
        """
        files_in_meta_file_match_files: no metadata file found.

        It cannot happen in normal circumstances: another validator for files
        is used before and raise a ValidationError for lack of metadata.

        Implemented here to have 100% branch coverage.
        """
        with self.assertRaises(AssertionError):
            artifact_utils.files_in_meta_file_match_files(
                "no_exist", deb822.Dsc, {}
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common functions for user input."""

import shutil
import subprocess
import sys
import tempfile
from collections.abc import Callable
from pathlib import Path

import yaml


class InvalidData(ValueError):
    """Exception raised when the user entered invalid data."""


class EditorInteractionError(RuntimeError):
    """
    Exception raised when interaction with the editor failed.

    This could be the editor not being found, or a failure to open the edited
    file.
    """


class YamlEditor[T]:
    """Let the user edit a data structure as YAML in a text editor."""

    def __init__(
        self,
        value: T,
        base_filename: str = "edit-",
        help_text: list[str] | None = None,
    ) -> None:
        """
        Set up editing a value.

        :param value: the initial value to edit
        :param base_filename: a string used as a prefix for constructing the
            temporary file name for the editor

        The edited value will be stored in self.value, and can be edited
        multiple times to allow downstream code to perform further validation.
        """
        self.initial = value
        self.value: T = value
        self.help_text = help_text or []

        # Reserve a file name to use for the editor
        with tempfile.NamedTemporaryFile(
            prefix=base_filename, suffix=".tmp.yaml", delete=False
        ) as editor_file:
            self.editor_file = Path(editor_file.name)

    def open_editor(self) -> None:
        """Open an editor on the given file."""
        editor = shutil.which("sensible-editor")
        if editor is None:
            raise EditorInteractionError(
                "sensible-editor not found:"
                " you may need to `apt install sensible-utils`"
            )
        subprocess.run([editor, self.editor_file.as_posix()])

    def read_edited(self) -> None:
        """
        Read the edited file, setting self.edited.

        If the edited file is empty, it sets self.edited to an empty valid
        value.

        :raises InvalidData: if the edited file is not valid YAML or is a value
            of an unexpected type
        :raises EditorInteractionError: if the edited file cannot be opened
        """
        try:
            with self.editor_file.open() as yaml_file:
                try:
                    contents = yaml.safe_load(yaml_file)
                except (yaml.YAMLError, yaml.scanner.ScannerError) as exc:
                    raise InvalidData(f"Invalid YAML: {exc}")
        except OSError as exc:
            raise EditorInteractionError(
                f"Cannot open edited file {self.editor_file}: {exc}"
            )

        if contents is None:
            # Empty YAML is normalized to empty data
            self.value = self.initial.__class__()
        elif not isinstance(contents, self.initial.__class__):
            raise InvalidData(
                f"Data must be a {self.initial.__class__.__name__} or empty"
            )
        else:
            self.value = contents

    def message(self, message: str) -> None:
        """Write a message to the user."""
        print(message)

    def confirm(self, prompt: str) -> bool:
        """Ask the user a yes/no question."""
        while True:
            res = input(prompt + " (y/n)> ")
            if res.strip().lower() == "y":
                return True
            elif res.strip().lower() == "n":
                return False

    def edit(self) -> bool:
        """
        Let the user edit a data structure as YAML in a text editor.

        :return: True if editing was successful, False if aborted.
        """
        # Write the original contents
        with self.editor_file.open("w") as out:
            for line in self.help_text:
                print("#", line, file=out)
            if self.help_text:
                print(file=out)
            yaml.safe_dump(data=self.value, sort_keys=False, stream=out)

        while True:
            self.open_editor()
            try:
                self.read_edited()
            except InvalidData as e:
                self.message(f"Error reading metadata: {e}")
                answer = self.confirm("Do you want to retry the same edit?")

                if not answer:
                    self.message(f"Edits left in {self.editor_file}")
                    return False
            else:
                self.cleanup()
                return True

    def edit_apply_loop[ResultT, ExceptionT: Exception](
        self,
        *,
        apply: Callable[[T], ResultT],
        apply_exception_class: type[ExceptionT],
        print_exception: Callable[[ExceptionT], None],
        content_description: str,
    ) -> ResultT | None:
        """
        Edit YAML, parse it, apply it, and allow retrying on error.

        :param apply: Function applying the parsed value and returning a result
        :param apply_exception_class: Exception raised by apply() if applying
          fails
        :param print_exception: Callable that prints the exception to stderr
        :param content_description: Description of the content being edited
        :return: Result of ``apply`` on success, else None
        """
        while True:
            if not self.edit():
                return None

            new = self.value
            if new == self.initial:
                return None

            try:
                return apply(new)
            except apply_exception_class as error:
                print(
                    f"Error setting {content_description}:\n", file=sys.stderr
                )
                print_exception(error)
            except ValueError as error:
                print(f"Error parsing {content_description}\n", file=sys.stderr)
                print(error, file=sys.stderr)

            if not self.confirm("\nDo you want to retry the same edit?"):
                return None

    def cleanup(self) -> None:
        """Delete the temporary file we created."""
        self.editor_file.unlink(missing_ok=True)
